# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
import platform
from qtpy import QtCore
from libopensesame.exceptions import ItemDoesNotExist
from libqtopensesame.widgets.base_widget import BaseWidget
from libopensesame import plugins, misc
from libqtopensesame.misc.markdown_parser import MarkdownParser
from libqtopensesame.misc import display
from libqtopensesame.misc.translate import translation_context
from qtpy.QtWebEngineWidgets import QWebEngineView as WebView
from qtpy.QtWebEngineWidgets import QWebEnginePage as WebPage
_ = translation_context('webbrowser', category='core')


# These urls are viewed internally in the browser component. All other urls are
# opened in an external browser.
INTERNAL_URLS = []


class SmallWebview(WebView):
    """A wrapper around QWebView too override the sizeHint, which prevents the
    browser from resizing to small sizes.
    """
    def sizeHint(self):
        """Gives a size hint.

        Returns
        -------
        A QSize
        """
        return QtCore.QSize(100, 100)


class SmallWebpage(WebPage):
    """A wrapper around QWeb(Engine)Page, to overload the
    acceptNavigationRequest function, which determines what needs to be done
    after a link is clicked.
    """
    def acceptNavigationRequest(self, *args):
        """Handles navigation requests to the browser, which can originate from
        link clicks, or other sources. the arguments and their order differ
        with the web browser backend being used, which can be QtWebKit or the
        newer QtWebEngine.

        Parameters
        ----------
        url: QtCore.QUrl
        navtype: QtWebEngine.NavigationType
        isMainFrame: bool
        """
        # Check if the first argument is a QUrl. If so, then
        # QWebEngine is used, if not, then QWebKit must be used. This way, the
        # order # of received arguments can be determined.
        if isinstance(args[0], QtCore.QUrl):
            url, navtype, isMainFrame = args
        else:
            frame, request, navtype = args
            url = request.url()
        if navtype == self.NavigationType.NavigationTypeLinkClicked:
            url = url.toString()
            if url.startswith('opensesame://'):
                self.parent().command(url)
                return False
            if url.startswith('new:'):
                self.parent().main_window.tabwidget.open_browser(url[4:])
                return False
            for internal_url in INTERNAL_URLS:
                if url.startswith(internal_url):
                    self.parent().load(url)
                    return False
            misc.open_url(url)
            return False
        return super().acceptNavigationRequest(*args)


class Webbrowser(BaseWidget):
    """A browser widget used to display online and offline help pages.
    
    Parameters
    ----------
    main_window: QtOpenSesame, optional
    """
    def __init__(self, main_window):
        super().__init__(main_window, ui='widgets.webbrowser_widget')
        self._current_url = ''
        self._cache = None
        self.ui.webview = SmallWebview(self)
        # Set webpage which handles link clicks
        webpage = SmallWebpage(self)
        self.ui.webview.setPage(webpage)
        # Touch events are enabled by default, and this has the effect that
        # touch events are broken for all other widgets once the webbrowser has
        # been used. This affects at least Ubuntu 15.05.
        self.ui.webview.setAttribute(QtCore.Qt.WA_AcceptTouchEvents, False)
        self.ui.webview.loadProgress.connect(self.update_progressbar)
        self.ui.webview.loadStarted.connect(self.load_started)
        self.ui.webview.loadFinished.connect(self.load_finished)
        self.ui.webview.urlChanged.connect(self.url_changed)
        self.ui.layout_main.addWidget(self.ui.webview)
        self.ui.button_back.clicked.connect(self.ui.webview.back)
        self.ui.button_osdoc.clicked.connect(self.open_osdoc)
        self.ui.button_forum.clicked.connect(self.open_forum)
        self.main_window.theme.apply_theme(self)
        self.markdown_parser = MarkdownParser(self)

    def load(self, url, tmpl=None):
        """Loads a webpage.

        Parameters
        ----------
        url
            The url to load.
        tmpl
            A template to be wrapped around the html in the case of Markdown
            files.
        """
        if isinstance(url, QtCore.QUrl):
            url = url.toString()
        if url.endswith('.md') and not url.startswith('http://') \
                and not url.startswith('https://'):
            self.ui.top_widget.hide()
            self.load_markdown(
                safe_read(url), url=os.path.basename(url),
                tmpl=tmpl
            )
            return
        self.ui.top_widget.show()
        self._current_url = url
        self.ui.webview.load(QtCore.QUrl(url))

    def load_markdown(self, md, url=None, tmpl=None):
        """Loads a Markdown text string.

        Parameters
        ----------
        md
            A Markdown text string.
        url
            The url to load.
        tmpl
            A template to be wrapped around the html.
        """
        if url is None:
            url = 'untitled'
        url = QtCore.QUrl('http://opensesame.app.cogsci.nl/%s' % url)
        self.ui.top_widget.hide()
        html = self.markdown_parser.to_html(md)
        if tmpl is not None:
            html = tmpl % {'body': html}
        self.ui.webview.setHtml(html, baseUrl=url)
        self.ui.webview.setZoomFactor(display.display_scaling)

    def load_finished(self, ok):
        """Hides the statusbar to indicate that loading is finished."""
        self.ui.label_load_progress.setText(_('Done'))
        self.ui.webview.setZoomFactor(display.display_scaling)

    def update_progressbar(self, progress):
        """Updates the progressbar to indicate the load progress.

        Parameters
        ----------
        progress
            The load progress.
        """
        self.ui.label_load_progress.setText('%d%%' % progress)

    def load_started(self):
        """Shows the statusbar to indicate that loading has started."""
        self.ui.label_load_progress.setText(_('Loading …'))

    def open_osdoc(self):
        """Opens osdoc.cogsci.nl."""
        self.load('http://osdoc.cogsci.nl/')

    def open_forum(self):
        """Opens forum.cogsci.nl."""
        self.load('http://forum.cogsci.nl/')

    def url_changed(self, url):
        """Updates the url bar.

        Parameters
        ----------
        url
            A url string.
        """
        self.ui.edit_url.setText(url.toString())

    def command(self, cmd):
        """Processes commands that are embedded in urls to trigger actions and
        events.

        Parameters
        ----------
        cmd : str
            A command string, such as 'action.save'.
        """
        cmd = cmd[13:]
        # This is quite a hacky workaround for Windows. The file paths are
        # automatically transformed into a Unix-like slashforward format.
        # Windows therefore cannot find the paths anymomre. To fix this, we
        # insert a colon, and normpath it.
        if platform.system() == 'Windows':
            _cmd = os.path.normpath(cmd[0] + ':' + cmd[1:])
        else:
            _cmd = cmd
        if os.path.exists(_cmd):
            self.main_window.open_file(path=_cmd, add_to_recent=False)
            return
        cmd = cmd.split('.')
        if len(cmd) == 2 and cmd[0] == 'action':
            try:
                action = getattr(self.main_window.ui, 'action_%s' % cmd[1])
            except:
                self.notify('Invalid action: %s' % cmd[1])
                return
            action.trigger()
            return
        if len(cmd) == 2 and cmd[0] == 'event':
            self.main_window.extension_manager.fire(cmd[1])
            return
        if len(cmd) > 2 and cmd[0] == 'item':
            try:
                self.experiment.items[cmd[1]].open_tab(phase=cmd[2],
                                                       args=cmd[3:])
            except ItemDoesNotExist:
                # The most likely case in which an item does not exist is when
                # there is an error from an Unknown item. We don't want to
                # give an error when participants click on this.
                pass
            return
        if len(cmd) > 1 and cmd[0] == 'help':
            if len(cmd) == 2:
                self.main_window.ui.tabwidget.open_help(cmd[1])
            elif len(cmd) == 3 and cmd[1] in ['extension', 'plugin']:
                if cmd[1] == 'extension':
                    folder == self.extension_manager[cmd[2]].folder
                else:
                    folder == self.plugin_manager[cmd[2]].folder
                path = os.path.join(folder, cmd[2] + '.md')
                self.load(path)
            return


# Alias for backwards compatibility
webbrowser = Webbrowser



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import DeviceError
from libopensesame.oslogging import oslogger
import serial
import os


class LibSrbox:

    """If you insert the srbox plugin at the start of your experiment, an
    instance of SRBOX automatically becomes part of the experiment
    object and
    can be accessed within an inline_script item as SRBOX.

    __Important note1:__

    If you do not specify a device, the plug-in will try to autodetect
    the
    SR Box port. However, on some systems this freezes the experiment, so
    it is better to explicitly specify a device.

    __Important note 2:__

    You
    need to call [srbox.start] to put the SR Box in sending mode,
    before
    calling [srbox.get_button_press] to collect a button press.

    __Example:__
    ~~~ .python
    t0 = clock.time()
    srbox.start()
    button, t1 = srbox.get_button_press(allowed_buttons=[1, 2],
                                        require_state_change=True)
    if button == 1:
        response_time = t1 - t0
    print(f'Button 1 was pressed in {response_time} ms!')
    srbox.stop()
    ~~~
    [TOC]
    """
    # The PST sr box only supports five buttons, but some of the VU boxes use
    # higher button numbers
    BUTTON1 = int('11111110', 2)
    BUTTON2 = int('11111101', 2)
    BUTTON3 = int('11111011', 2)
    BUTTON4 = int('11110111', 2)
    BUTTON5 = int('11101111', 2)
    BUTTON6 = int('11011111', 2)
    BUTTON7 = int('10111111', 2)
    BUTTON8 = int('01111111', 2)
    BYTEMASKS = [
        BUTTON1,
        BUTTON2,
        BUTTON3,
        BUTTON4,
        BUTTON5,
        BUTTON6,
        BUTTON7,
        BUTTON8
    ]

    def __init__(self, experiment, dev=None):
        """Constructor. An SRBOX object is created automatically by the SRBOX
        plug-in, and you do not generally need to call the constructor
        yourself.

        Parameters
        ----------
        experiment : experiment
            An Opensesame experiment.
        """
        self.experiment = experiment
        self._srbox = None
        self._started = False

        # If a device has been specified, use it
        if dev not in (None, "", "autodetect"):
            try:
                self._srbox = serial.Serial(dev, timeout=0, baudrate=19200)
            except Exception as e:
                raise DeviceError(
                    "Failed to open device port '%s' in libsrbox: '%s'"
                    % (dev, e)
                )

        else:
            # Else determine the common name of the serial devices on the
            # platform and find the first accessible device. On Windows,
            # devices are labeled COM[X], on Linux there are labeled /dev/tty[X]
            if os.name == "nt":
                for i in range(1, 256):
                    try:
                        dev = "COM%d" % i
                        self._srbox = serial.Serial(
                            dev,
                            timeout=0,
                            baudrate=19200
                        )
                        break
                    except Exception as e:
                        self._srbox = None
                        pass
            elif os.name == "posix":
                for path in os.listdir("/dev"):
                    if path[:3] == "tty":
                        try:
                            dev = "/dev/%s" % path
                            self._srbox = serial.Serial(
                                dev,
                                timeout=0,
                                baudrate=19200
                            )
                            break
                        except Exception as e:
                            self._srbox = None
                            pass
            else:
                raise DeviceError(
                    u"libsrbox does not know how to auto-detect the SR Box on "
                    u"your platform. Please specify a device."
                )
        if self._srbox is None:
            raise DeviceError(
                u"libsrbox failed to auto-detect an SR Box. Please specify a "
                u"device."
            )
        oslogger.debug("using device %s" % dev)
        # Turn off all lights
        if self._srbox is not None:
            self._srbox.write(b'\x60')

    def send(self, ch):
        """Sends a single character to the SR Box. Send '\x60' to turn off all
        lights, '\x61' for light 1 on, '\x62' for light 2 on,'\x63' for lights
        1 and 2 on etc.

        Parameters
        ----------
        ch : str or bytes
            The character to send. If a `str` is passed, it is encoded to
            `bytes` using utf-8 encoding.
        """
        if isinstance(ch, str):
            ch = ch.encode('utf-8')
        self._srbox.write(ch)

    def start(self):
        """Turns on sending mode, so that the SR Box starts to send output.
        The SR Box must be in sending mode when you call
        [srbox.get_button_press].
        """
        if self._started:
            return
        # Write the start byte
        self._srbox.flushOutput()
        self._srbox.flushInput()
        self._srbox.write(b'\xA0')
        self._started = True

    def stop(self):
        """Turns off sending mode, so that the SR Box stops giving output."""
        if not self._started:
            return
        # Write the stop byte and flush the input
        self._srbox.flushOutput()
        self._srbox.flushInput()
        self._srbox.write(b'\x20')
        self._started = False

    def get_button_press(self, allowed_buttons=None, timeout=None,
                         require_state_change=False):
        """Collects a button press from the SR box.

        Parameters
        ----------
        allowed_buttons : list, NoneType, optional
            A list of buttons that are accepted or `None` to accept all
            buttons. Valid buttons are integers 1 through 8.
        timeout : int, float, NoneType, optional
            A timeout value in milliseconds or `None` for no timeout.
        require_state_change    Indicates whether already pressed button should be accepted
            (False), or whether only a state change from unpressed to pressed
            is accepted (True).

        Returns
        -------
        tuple
            A `(button_list, timestamp)` tuple. `button_list` is `None` if no 
            button was pressed (i.e. a timeout occurred).
        """
        if not self._started:
            raise DeviceError(
                'Please call srbox.start() before srbox.get_button_press()')
        t0 = self.experiment.clock.time()
        # Create a list of buttonr, bytemask tuples.
        bytemasks = []
        for buttonnr, bytemask in enumerate(self.BYTEMASKS):
            if allowed_buttons is None or buttonnr + 1 in allowed_buttons:
                bytemasks.append((buttonnr + 1, bytemask))
        inputbyte0 = None
        while True:
            # Get a valid character and convert it to a byte
            inputchar = self._srbox.read(1)
            if inputchar == b'':
                continue
            inputbyte1 = ord(inputchar)
            # Check for a timeout
            t1 = self.experiment.clock.time()
            if timeout is not None and t1 - t0 > timeout:
                break
            # To check for state changes, we need an old and a new state.
            # Therefore, on the first loop we don't do anything.
            if require_state_change and inputbyte0 is None:
                inputbyte0 = inputbyte1
                continue
            # Ignore no responses
            if not inputbyte1:
                inputbyte0 = inputbyte1
                continue
            if require_state_change:
                # If a state change is required, a button should be pressed now
                # but not before. Because there is only one state change at a
                # time, we can return right away.
                for buttonnr, bytemask in bytemasks:
                    if (
                            inputbyte1 | bytemask == 255 and
                            inputbyte0 | bytemask != 255
                    ):
                        return [buttonnr], t1
            else:
                # If no state change is required, only consider the last input
                # and create a list of all buttons that are pressed.
                button_list = []
                for button_nr, bytemask in bytemasks:
                    if inputbyte1 | bytemask == 255:
                        button_list.append(button_nr)
                if button_list:
                    return button_list, t1
            inputbyte0 = inputbyte1
        return None, t1

    def close(self):
        """Closes the connection to the srbox. This is done automatically by
        the SRBOX plugin when the experiment finishes. {nodoc}
        """
        self._srbox.close()
        self._started = False


# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import sys
from libqtopensesame.misc.config import cfg
from libopensesame.oslogging import oslogger
from pyqode.core.managers import BackendManager
from pyqode.core.backend import server
from pyqode.core import panels, modes
from pyqode.core.api import (
    CodeEdit,
    Panel,
    SyntaxHighlighter,
    ColorScheme
)


class TextSH(SyntaxHighlighter):

    def highlight_block(self, text, user_data):

        pass


class TextCodeEdit(CodeEdit):

    mimetypes = ['text/x-plain', 'text/x-log', 'text/plain']
    language = 'text'

    def __init__(self, parent=None):

        # PsychoPy deletes the _ built-in
        if '_' not in __builtins__:
            oslogger.warning('re-installing missing gettext built-in')
            import gettext
            gettext.NullTranslations().install()
        _reset_stylesheet = self._reset_stylesheet
        self._reset_stylesheet = lambda: None
        super(TextCodeEdit, self).__init__(parent)
        self._backend = BackendManager(self)
        self.backend.start(
            server.__file__,
            sys.executable,
            reuse=True,
            share_id='text'
        )
        # append panels
        self.panels.append(
            panels.SearchAndReplacePanel(),
            Panel.Position.BOTTOM
        )
        self.panels.append(panels.FoldingPanel())
        self.panels.append(panels.LineNumberPanel())
        # append modes
        self.modes.append(modes.AutoCompleteMode())
        self.modes.append(modes.ExtendedSelectionMode())
        self.modes.append(modes.CaseConverterMode())
        self.modes.append(modes.FileWatcherMode())
        self.modes.append(modes.CaretLineHighlighterMode())
        self.modes.append(modes.RightMarginMode())
        self.modes.append(
            TextSH(
                self.document(),
                color_scheme=ColorScheme(cfg.pyqode_color_scheme)
            )
        )
        self.modes.append(modes.ZoomMode())
        self.modes.append(modes.CodeCompletionMode())
        self.modes.append(modes.AutoIndentMode())
        self.modes.append(modes.IndenterMode())
        self.modes.append(modes.SymbolMatcherMode())
        self.panels.append(panels.EncodingPanel(), Panel.Position.TOP)
        self.panels.append(panels.ReadOnlyPanel(), Panel.Position.TOP)
        self._reset_stylesheet = _reset_stylesheet

    def setPlainText(self, txt, mime_type='', encoding=''):

        super(TextCodeEdit, self).setPlainText(
            txt,
            mime_type,
            encoding
        )

    def clone(self):

        clone = self.__class__(parent=self.parent())
        return clone



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import InvalidFormScript
from libopensesame.widgets._button import Button


class Checkbox(Button):

    r"""The `Checkbox` widget is a checkable box accompanied by a string of
    text.

    __Example (OpenSesame script):__

    ~~~
    widget 0 0 1 1 checkbox
    group="group" text="Option 1"
    widget 0 1 1 1 checkbox group="group"
    text="Option 2"
    ~~~

    __Example (Python):__

    ~~~ .python
    form = Form()
    checkbox1 = Checkbox(text='Option 1', group='group')
    checkbox2 =
    Checkbox(text='Option 2', group='group')
    form.set_widget(checkbox1, (0,0))
    form.set_widget(checkbox2, (0,1))
    ~~~

    [TOC]
    """
    def __init__(self, form, text=u'checkbox', frame=False, group=None,
                 checked=False, click_accepts=False, var=None):
        r"""Constructor to create a new `Checkbox` object. You do not generally
        call this constructor directly, but use the `Checkbox()` factory
        function, which is described here: [/python/common/]().

        Parameters
        ----------
        form : form
            The parent form.
        text : str, unicode, optional
            Checkbox text.
        frame : bool, optional
            Indicates whether a frame should be drawn around the widget.
        group : str, unicode, NoneType, optional
            If a group is specified, checking one checkbox from the group will
            uncheck all other checkboxes in that group. Checkboxes that are
            part of a group cannot be unchecked, except by clicking on another
            checkbox in the group. The `group` keyword also affects how
            variables are stored (see the `var` keyword).
        checked : bool, optional
            The initial checked state of the checkbox.
        click_accepts : bool, optional
            Indicates whether a click press should accept and close the form.
        var : str, unicode, NoneType, optional
            The name of the experimental variable that should be used to log
            the widget status. This variable will contain a semi-colon
            separated list of the text of all checked checkboxes in the same
            group (falling back to checkbox1, etc. if no text has been
            specified), or 'no' if no checkbox in the group is checked.  For
            the purpose of the variable, all checkboxes that are not part of a
            group are placed in the same group. For more information about the
            use of response variables in forms, see the form documentation
            page.
        """
        if isinstance(checked, str):
            checked = checked == u'yes'
        if isinstance(click_accepts, str):
            click_accepts = click_accepts == u'yes'
        Button.__init__(self, form, text, frame=frame, center=False)
        self.type = u'checkbox'
        self.group = group
        self.box_pad = self.x_pad
        self.x_pad += self.x_pad + self.box_size
        self.var = var
        self.click_accepts = click_accepts
        self.checked = checked

    def on_mouse_click(self, pos):
        r"""Is called whenever the user clicks on the widget. Toggles the state
        of the checkbox.

        Parameters
        ----------
        pos : tuple
            An (x, y) coordinate tuple.
        """
        if self.group is not None:
            # If the checkbox is part of a group than checking it will uncheck
            # all other checkboxes in the group, and check the current one
            for widget in self.form.widgets:
                if widget is not None and widget.type == u'checkbox' and \
                        widget.group == self.group and self.group is not None:
                    widget.set_checked(False)
            self.set_checked(True)
        else:
            # If the checkbox is not part of a group then checking it will
            # toggle its check status
            self.set_checked(not self.checked)
        if self.click_accepts:
            return self.text

    def _init_canvas_elements(self):

        x, y, w, h = self.rect
        self._checked_element = self.form.theme_engine.box(x+self.box_pad,
                                                           y+self.y_pad, checked=True)
        self._unchecked_element = self.form.theme_engine.box(x+self.box_pad,
                                                             y+self.y_pad, checked=False)
        self.canvas.add_element(self._checked_element)
        self.canvas.add_element(self._unchecked_element)
        Button._init_canvas_elements(self)
        self.set_checked(self.checked)

    def set_checked(self, checked=True):
        r"""Sets the checked status of the checkbox.

        Parameters
        ----------
        checked : bool, optional
            The checked status.
        """
        self._checked_element.visible = checked
        self._unchecked_element.visible = not checked
        self.checked = checked
        self.set_var(checked)

    def set_var(self, val, var=None):
        r"""Sets an experimental variable.

        Parameters
        ----------
        val : str, unicode
            A value.
        var : str, unicode, NoneType, optional
            A variable name, or `None` to use widget default.
        """
        if var is None:
            var = self.var
        if var is None:
            return
        # Set the response variable
        l_val = []
        # When this function is called via the constructor, the checkbox is not
        # yet part of the form. Therefore, we need to add it explicitly to the
        # widget list.
        widget_list = self.form.widgets[:]
        if self not in self.form.widgets:
            widget_list += [self]
        for widget in widget_list:
            # Only consider checkbox widgets that use the same variable as
            # the current checkbox
            if (
                    widget is None or
                    widget.type != u'checkbox' or
                    widget.var != self.var
            ):
                continue
            if widget.group != self.group and self.group is not None:
                raise InvalidFormScript('All checkbox widgets without a '
                                        'group or within the same group '
                                        'should have the same variable.')
            # Ignore the checkbox if it isn't checked
            if not widget.checked and widget.checked != u'yes':
                continue
            # The variable is normally the text on the checkbox. However, if
            # no text has been specified, then we use checkbox1, etc. as text
            text = safe_decode(widget.text).strip()
            if not text:
                text = 'checkbox{}'.format(widget_list.index(widget))
            l_val.append(text)
        val = u';'.join(l_val)
        if val == u'':
            val = u'no'
        Button.set_var(self, val, var=var)


checkbox = Checkbox



# -*- coding:utf-8 -*-

"""
This file is part of openexp.

openexp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

openexp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with openexp.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.sketchpad_elements._base_element import BaseElement
from libopensesame.sketchpad_elements import Image as ImageRuntime


class Image(BaseElement, ImageRuntime):

    r"""An image element.
    See base_element for docstrings and function
    descriptions.
    """
    @classmethod
    def mouse_press(cls, sketchpad, pos):

        from libqtopensesame.widgets import pool_widget
        _file = pool_widget.select_from_pool(sketchpad.main_window,
                                             parent=sketchpad._edit_widget)
        if _file == u'':
            return None
        properties = {
            u'x': pos[0],
            u'y': pos[1],
            u'file': _file,
            u'scale': sketchpad.current_scale(),
            u'rotation': sketchpad.current_rotation(),
            u'center': sketchpad.current_center(),
            u'show_if': sketchpad.current_show_if()
        }
        e = Image(sketchpad, properties=properties)
        return e

    @staticmethod
    def requires_center():
        return True

    @staticmethod
    def requires_scale():
        return True

    @staticmethod
    def requires_rotation():
        return True


# Alias for backwards compatibility
image = Image



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.widgets.tree_overview import TreeOverview


class TreeOverviewAdapter(TreeOverview):
    r"""Extends the tree overview so that it treats coroutines as sequences,
    and supports start and end times.
    """
    
    def __init__(self, coroutines, main_window, overview_mode=True):
        self.coroutines = coroutines
        super().__init__(main_window, overview_mode=overview_mode)

    def text_edited(self, treeitem, col):
        # Renames and run-if edits are handled by the original tree overview,
        # which needs to be tricked into thinking that the coroutines item is
        # a sequence.
        if col in (0, 1):
            self.coroutines.item_type = u'sequence'
            super().text_edited(treeitem, col)
            self.coroutines.item_type = u'coroutines'
            return
        # Don't allow editing the times of the top-level coroutines
        if treeitem.parent() is None:
            self.itemChanged.disconnect()
            treeitem.setText(col, u'')
            self.itemChanged.connect(self.text_edited)
            return
        # Change the start/ end time
        parent_item_name, ancestry = treeitem.ancestry()
        parent_item_name, index = self.parent_from_ancestry(ancestry)
        _time = treeitem.text(col).strip()
        if _time == u'':
            _time = 0
        item_name, start_time, end_time, cond, enabled = \
            self.coroutines.schedule[index]
        if col == 2:
            start_time = _time
        elif col == 3:
            # In the GUI one-shot coroutines are not associated with an
            # end time. In the runtime, their end time is equal to the start
            # time.
            if self.coroutines.is_oneshot_coroutine(item_name):
                _time = u''
                end_time = start_time
            else:
                end_time = _time
        else:
            raise TypeError(u'col should be in 0-3 range')
        self.coroutines.schedule[index] = \
            item_name, start_time, end_time, cond, enabled
        self.itemChanged.disconnect()
        treeitem.setText(col, str(_time))
        self.itemChanged.connect(self.text_edited)
        self.text_change.emit()



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from libqtopensesame.widgets.base_widget import BaseWidget
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from openexp import backend
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'backend_settings', category=u'core')


class BackendSettings(BaseWidget):

    def __init__(self, main_window):

        super().__init__(main_window, ui=u'widgets.backend_settings')
        self.tab_name = u'__backend_settings__'
        for backend_type in backend._backend_types:
            try:
                _backend = backend.get_backend_class(self.experiment,
                                                     backend_type)
            except:
                _backend = None
            layout = getattr(self.ui, u'layout_%s' % backend_type)
            label = getattr(self.ui, u'label_%s' % backend_type)
            # Horribly ugly way to clear the previous settings
            while layout.count() > 1:
                w = layout.itemAt(1)
                layout.removeItem(w)
                w.widget().hide()
            if _backend is None:
                label.setText(_(u"Failed to load backend"))
            elif not hasattr(_backend, u"settings") or _backend.settings == \
                    None:
                label.setText(_(u"No settings for %s") % _backend.__name__)
            else:
                label.setText(_(u"Settings for %s:") % _backend.__name__)
                layout.addWidget(settings_widget(self.main_window,
                                                 _backend.settings))


class SettingsEdit(QtWidgets.QLineEdit, BaseSubcomponent):

    """An edit widget for a single variable"""
    def __init__(self, main_window, var, val):
        """
        Constructor

        Arguments:
        experiment -- the experiment
        var -- the variable name
        val -- the variable value

        Keywords arguments:
        parent -- parent QWidget (default=None)
        """
        super().__init__()
        self.setup(main_window)
        self.setText(safe_decode(val))
        self.var = var
        self.editingFinished.connect(self.apply_setting)

    def apply_setting(self):
        """Apply changes"""
        self.experiment.var.set(
            self.var, self.experiment.syntax.sanitize(self.text()))


class SettingsWidget(BaseWidget):

    """A widget containing a number of settings"""
    def __init__(self, main_window, settings):
        """
        Constructor

        Arguments:
        experiment -- the experiment
        settings -- the settings dictionary

        Keywords arguments:
        parent -- parent QWidget (default=None)
        """
        super().__init__(main_window)
        self.settings = settings
        self.layout = QtWidgets.QFormLayout(self)
        self.layout.setFieldGrowthPolicy(
            QtWidgets.QFormLayout.FieldsStayAtSizeHint)
        self.setLayout(self.layout)
        for var, desc in settings.items():
            if var in self.experiment.var:
                val = self.experiment.var.get(var)
            else:
                val = desc[u"default"]
            label = QtWidgets.QLabel()
            label.setText(
                u"%(name)s<br /><small><i>%(description)s</i></small>" % desc)
            label.setTextFormat(QtCore.Qt.RichText)
            edit = settings_edit(self.main_window, var, val)
            self.layout.addRow(label, edit)


# Alias for backwards compatibility
backend_settings = BackendSettings
settings_edit = SettingsEdit
settings_widget = SettingsWidget



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy.QtWidgets import QLineEdit
from qtpy.QtGui import QColor, QBrush, QPainter, QPen, QFont
from qtpy.QtCore import Qt
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.misc.translate import translation_context
_ = translation_context('conditional_expression', category='core')


class ConditionalExpression(QLineEdit, BaseSubcomponent):
    """A line-edit widget for conditional expressions. Automatically rewrites
    False and True expressions to be more readable, and adds a color hit.
    """
    
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setup(parent)
        self.textChanged.connect(self._set_conditional_expression)
        self._verb = ''
        self._font = QFont('Roboto Condensed')
        self._pen = QPen(QColor('#78909c'))
        self._margin = 16
        self._width = None
        self._flags = Qt.AlignLeft | Qt.AlignVCenter
        self._annotations = {'True': 'Always', 'False': 'Never'}
        
    @property
    def verb(self):
        return self._verb
    
    @verb.setter
    def verb(self, value):
        self._verb = value
        self._annotations = {'True': f'Always {value}',
                             'False': f'Never {value}'}
    
    def _set_conditional_expression(self, cond):
        self.textChanged.disconnect(self._set_conditional_expression)
        fixed_cond = self.experiment.syntax.fix_conditional_expression(cond)
        if cond != fixed_cond:
            self.setText(fixed_cond)
        else:
            clean_cond = cond.strip().lower()
            color = None
            if clean_cond in ('always', 'true'):
                self.setText('True')
                color = 'green'
            elif clean_cond in ('never', 'false'):
                self.setText('False')
            if color is not None:
                self.setStyleSheet(f'color: {color}')
            else:
                self.setStyleSheet(None)
        self.textChanged.connect(self._set_conditional_expression)

    def paintEvent(self, event):
        """Draws annotations to describe the conditional expressions for the
        user.
        """
        super().paintEvent(event)
        painter = QPainter(self)
        data = self.text()
        annotation = self._annotations.get(data, None)
        if annotation is None:
            return
        # The first time that we draw an annotation, we determine how far from
        # the left we need to draw it. This depends on the margin and the size
        # of the text
        event_rect = event.rect()
        if self._width is None:
            rect = painter.boundingRect(event_rect, 0, data)
            self._width = rect.right() + self._margin
        event_rect.setLeft(self._width)
        painter.setFont(self._font)
        painter.setPen(self._pen)
        painter.drawText(event_rect, self._flags, annotation)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.sketchpad import Sketchpad


class Feedback(Sketchpad):

    description = u'Provides feedback to the participant'

    def reset(self):
        """See item."""
        super().reset()
        self.var.reset_variables = u'yes'

    def prepare(self):
        """Prepares the item."""
        pass

    def run(self):
        """Runs the item."""
        super().prepare()
        super().run()
        if self.var.reset_variables == u'yes':
            self.experiment.reset_feedback()

    def coroutine(self):
        """See coroutines plug-in."""
        super().prepare()
        yield
        self.set_item_onset(self.canvas.show())


# Alias for backwards compatibility
feedback = Feedback



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets, QtGui
from libqtopensesame.widgets.base_widget import BaseWidget
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'color_edit', category=u'core')


class ColorEdit(BaseWidget):

    r"""A colorpicker widget that emulates a QLineEdit."""
    textChanged = QtCore.Signal('QString')
    textEdited = QtCore.Signal('QString')

    def __init__(self, parent):
        r"""Constructor.

        Parameters
        ----------
        main_window : qtopensesame
            The main-window object.
        """
        super().__init__(parent)
        self.edit = QtWidgets.QLineEdit()
        self._parent = None
        self.edit.setSizePolicy(QtWidgets.QSizePolicy.Minimum,
                                QtWidgets.QSizePolicy.Minimum)
        self.edit.editingFinished.connect(self.apply)
        self.editingFinished = self.edit.editingFinished
        self.button = QtWidgets.QPushButton()
        self.button.setIconSize(QtCore.QSize(16, 16))
        self.button.clicked.connect(self.colorpicker)
        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.edit)
        layout.addWidget(self.button)
        self.setLayout(layout)

    def colorpicker(self):
        r"""Picks a color with the colorpicker dialog."""
        _color = QtWidgets.QColorDialog.getColor(QtGui.QColor(u'white'),
                                                 self._parent, _(u'Pick a color'))
        if not _color.isValid():
            return
        self.setText(_color.name())
        self.apply()

    def text(self):
        r"""Gets text (emulate QLineEdit behavior).

        Returns
        -------
        QString
            A color text.
        """
        return self.edit.text()

    def setText(self, s):
        """
        dsc:
                Sets text (emulate QLineEdit behavior).

        arguments:
                s:
                        desc:	Text.
                        type:	unicode
        """
        self.edit.setText(safe_decode(s))

    def apply(self):
        r"""Emits a 'set_color' signal to indicate that a color has been
        picked.
        """
        self.textChanged.emit(self.text())
        self.textEdited.emit(self.text())

    def initialize(self, experiment=None, color=None, parent=None):
        r"""Initializes the widget. This is necessary to apply the theme and
        give the fields initial values.

        Parameters
        ----------
        experiment : experiment, NoneType, optional
            The experiment object or None if it is already available via the
            base_component property.
        color : unicode, NoneType, optional
        parent, optional
            A parent QWidget.
        """
        if parent is not None:
            self._parent = parent
        else:
            self._parent = self.main_window
        if experiment is None:
            experiment = self.experiment
        if color is None:
            color = experiment.var.get(u'foreground', _eval=False)
        self.setText(color)
        self.button.setIcon(self.theme.qicon(u'os-color-picker'))


# Alias for backwards compatibility
color_edit = ColorEdit



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from pygame.locals import *
from openexp._mouse.mouse import Mouse
from openexp._mouse.legacy import Legacy
from openexp.backend import configurable
import pygame
from expyriment import stimuli
from expyriment.misc.geometry import coordinates2position as c2p
from openexp._coordinates.xpyriment import Xpyriment as XpyrimentCoordinates


class Xpyriment(XpyrimentCoordinates, Legacy):

    r"""This is a mouse backend built on top of PyGame, adapted for Expyriment.
    For function specifications and docstrings, see `openexp._mouse.mouse`.
    """
    settings = {}

    def __init__(self, experiment, **resp_args):

        Mouse.__init__(self, experiment, **resp_args)
        XpyrimentCoordinates.__init__(self)

    @configurable
    def get_click(self):

        buttonlist = self.buttonlist
        timeout = self.timeout
        pygame.mouse.set_visible(self.visible)
        start_time = pygame.time.get_ticks()
        time = start_time
        while True:
            time = pygame.time.get_ticks()
            # Process the input
            for event in pygame.event.get():
                if event.type == KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        self.experiment.pause()
                        continue
                    pygame.event.post(event)
                if event.type == MOUSEBUTTONDOWN:
                    if buttonlist is None or event.button in buttonlist:
                        pygame.mouse.set_visible(self._cursor_shown)
                        return event.button, self.from_xy(event.pos), time
            if timeout is not None and time - start_time >= timeout:
                break
        pygame.mouse.set_visible(self._cursor_shown)
        return None, None, time


# Non PEP-8 alias for backwards compatibility
xpyriment = Xpyriment



"""A simple notepad to document your experiment. This plug-in does nothing."""

category = 'Miscellaneous'
icon = 'notes'
controls = [
    {
        "type": "editor",
        "var": "note",
        "label": "Note",
        "name": "note_widget",
        "tooltip": "Type your note here",
        "info": None,
        "min_width": None,
        "prefix": "",
        "suffix": "",
        "left_label": "min.",
        "right_label": "max.",
        "syntax": False
    }
]



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._canvas._element.element import Element


class Image(Element):

    def __init__(self, canvas, fname, center=True, x=None, y=None, scale=None,
                 rotation=None, **properties):

        x, y = canvas.none_to_center(x, y)
        self._image_size = None
        self._shapely_polygon = None
        Element.__init__(
            self, canvas,
            fname=fname,
            center=center,
            x=x,
            y=y,
            scale=scale,
            rotation=rotation,
            **properties
        )

    def _size(self):

        if self._image_size is None:
            from PIL import Image
            w, h = Image.open(self.fname).size
            if self.scale is not None:
                w *= self.scale
                h *= self.scale
            self._image_size = w, h
        return self._image_size

    @property
    def rect(self):

        from PIL import Image

        im = Image.open(self.fname)
        w1, h1 = im.size
        if self.rotation is not None and self.rotation != 0:
            im = im.rotate(self.rotation, expand=True)
        w2, h2 = im.size
        im.close()
        dx = (w2-w1)/2
        dy = (h2-h1)/2
        if self.scale is not None:
            w2 *= self.scale
            h2 *= self.scale
            dx *= self.scale
            dy *= self.scale
        x, y = self.none_to_center(self.x, self.y)
        if self.center:
            return x-w2//2, y-h2//2, w2, h2
        return x-dx, y-dy, w2, h2

    def __contains__(self, xy):

        # Shapely is used to determine whether a point falls exactly within the
        # polygon. If shapely isn't available, we fall back to a simple bounding
        # box. The shapely polygon is stored for performance.
        try:
            from shapely.geometry import box, Point
            from shapely import affinity
        except ImportError:
            return Element.__contains__(self, xy)
        if self._shapely_polygon is None:
            w, h = self._size()
            x, y = self.none_to_center(self.x, self.y)
            if self.center:
                x -= w / 2
                y -= h / 2
            self._shapely_polygon = box(x, y, x + w, y + h)
            if self.rotation:
                self._shapely_polygon = affinity.rotate(
                    self._shapely_polygon,
                    self.rotation
                )
        return self._shapely_polygon.contains(Point(*xy))

    def _on_attribute_change(self, **kwargs):

        # When a change occurs the shapely line (if any) needs to be
        # redetermined in __contains__
        self._shapely_polygon = None
        Element._on_attribute_change(self, **kwargs)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from threading import Thread, Lock
import time


class Heartbeat(Thread):

    r"""A thread that sends regular heartbeats to the launch process (if any).
    A heartbeat is a transfer of the experiment workspace.
    """
    def __init__(self, exp, interval=1):
        r"""Constructor.

        Parameters
        ----------
        exp : experiment
            The experiment object.
        interval : float, int, optional
            The heartbeat interval in seconds.
        """
        super().__init__()
        self.exp = exp
        self.interval = interval
        self.lock = Lock()

    def run(self):
        r"""Runs the heartbeat loop."""
        while self.exp.running:
            time.sleep(self.interval)
            self.beat()

    def beat(self):
        r"""Sends a single heartbeat."""
        self.lock.acquire()
        self.exp.transmit_workspace(__heartbeat__=True)
        self.lock.release()


# Alias for backwards compatibility
heartbeat = Heartbeat



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
from libopensesame.widgets.themes.plain import Plain
from openexp import resources
from openexp.canvas_elements import Rect, Line, Image
from openexp._canvas._element.group import Group


class Gray(Plain):

    def __init__(self, form):

        super().__init__(form)
        self.box_checked_image = resources['widgets/gray/box-checked.png']
        self.box_unchecked_image = resources['widgets/gray/box-unchecked.png']

    def box(self, x, y, checked=False):

        return Image(
            self.box_checked_image if checked else self.box_unchecked_image,
            center=False, x=x, y=y).construct(self.canvas)

    def frame(self, x, y, w, h, style=u'normal'):

        elements = [
            Line(x+1, y+1, x+w-2, y+1, color=u'#babdb6'),
            Line(x+1, y+1, x+1, y+h-2, color=u'#babdb6'),
            Line(x+1, y+h-2, x+w-2, y+h-2, color=u'#555753'),
            Line(x+w-2, y+1, x+w-2, y+h-2, color=u'#555753')
        ]
        if style in (u'active', u'normal'):
            elements.insert(0, Rect(x, y, w, h, color=u'#2e3436'))
        if style == u'normal':
            elements.append(
                Rect(x+2, y+2, w-4, h-4, color=u'#888a85', fill=True))
        return Group(self.canvas, elements)


# Alias for backwards compatibility
gray = Gray



"""Initializes a new Quest staircase procedure"""

category = "Staircase"
controls = [
    {
        "type": "line_edit",
        "var": "t_guess",
        "label": "Estimated threshold",
        "info": "Used for initial test value"
    },
    {
        "type": "line_edit",
        "var": "t_guess_sd",
        "label": "Std. dev. of estimated threshold",
    },
    {
        "type": "line_edit",
        "var": "p_threshold",
        "label": "Target",
        "info": "Desired proportion of correct responses"
    },
    {
        "type": "line_edit",
        "var": "beta",
        "label": "\u03b2",
        "info": "Steepness of the Weibull psychometric function"
    },
    {
        "type": "line_edit",
        "var": "delta",
        "label": "\u03b4",
        "info": "Proportion of random responses at maximum stimulus intensity"
    },
    {
        "type": "line_edit",
        "var": "gamma",
        "label": "Chance level (\u03b3)",
    },
    {
        "type": "combobox",
        "var": "test_value_method",
        "label": "Method",
        "options": [
            "quantile",
            "mean",
            "mode"
        ],
        "tooltip": "Method to determine optimal test value"
    },
    {
        "type": "line_edit",
        "var": "min_test_value",
        "label": "Minimum test value",
    },
    {
        "type": "line_edit",
        "var": "max_test_value",
        "label": "Maximum test value",
    },
    {
        "type": "line_edit",
        "var": "var_test_value",
        "label": "Experimental variable for test value",
    },
    {
        "type": "line_edit",
        "var": "quest_name",
        "label": "Name",
        "info": "Use different names to run multiple independent Quest procedures"
    }
]



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import functools
from qtpy.QtWidgets import (QCheckBox,
                            QSpinBox,
                            QDoubleSpinBox,
                            QKeySequenceEdit,
                            QLineEdit,
                            QComboBox,
                            QFontComboBox,
                            QPlainTextEdit)
from qtpy.QtGui import QKeySequence, QFont
from libopensesame.oslogging import oslogger
from libqtopensesame.widgets.base_widget import BaseWidget
from libqtopensesame.misc.config import cfg


class BasePreferencesWidget(BaseWidget):

    def __init__(self, main_window, ui):
        r"""Constructor.

        Parameters
        ----------
        main_window
            A qtopensesame object.
        ui
            A ui file.
        """
        super().__init__(main_window, ui=ui)
        self._its_me = False
        self._before_init_widgets()
        self._init_widgets()
        self._after_init_widgets()
        self.extension_manager.register_extension(self)

    def _before_init_widgets(self):
        r"""Can be implemented to perform some actions before the rest of the
        widgets are initialized.
        """
        pass

    def _after_init_widgets(self):
        r"""Can be implemented to perform some actions after the rest of the
        widgets have been initialized.
        """
        pass

    def _init_widgets(self):
        r"""Introspects the widgets and automatically binds them to settings."""
        for name in dir(self.ui):
            if not name.startswith('cfg_'):
                continue
            setting = name[4:]
            if setting not in cfg:
                oslogger.warning('unknown setting {}'.format(setting))
            widget = getattr(self.ui, name)

            def change_setting(setting, value):
                if self._its_me:
                    return
                if isinstance(value, QKeySequence):
                    value = value.toString()
                elif isinstance(value, QFont):
                    value = value.family()
                elif isinstance(value, QLineEdit):
                    value = value.text()
                elif isinstance(value, QPlainTextEdit):
                    value = value.toPlainText()
                cfg[setting] = value
                self._its_me = True
                self.extension_manager.fire(
                    'setting_changed',
                    setting=setting,
                    value=value
                )
                self._its_me = False
            # The setting needs to be bound to the function because it is not
            # passed by the signal
            change_setting = functools.partial(change_setting, setting)
            # For QPlainTextEdit widgets, the widget itself also needs to be
            # bound, because the signal doesn't pass the widget
            change_plaintext_setting = functools.partial(change_setting,
                                                         widget)

            try:
                if isinstance(widget, QCheckBox):
                    widget.setChecked(cfg[setting])
                    widget.toggled.connect(change_setting)
                elif isinstance(widget, QLineEdit):
                    widget.setText(safe_decode(cfg[setting]))
                    widget.editingFinished.connect(
                        functools.partial(change_setting, widget))
                elif isinstance(widget, (QSpinBox, QDoubleSpinBox)):
                    widget.setValue(cfg[setting])
                    widget.valueChanged.connect(change_setting)
                elif isinstance(widget, QKeySequenceEdit):
                    widget.setKeySequence(QKeySequence(cfg[setting]))
                    widget.keySequenceChanged.connect(change_setting)
                elif isinstance(widget, QFontComboBox):
                    widget.setCurrentFont(QFont(cfg[setting]))
                    widget.currentFontChanged.connect(change_setting)
                elif isinstance(widget, QComboBox):
                    widget.setCurrentText(cfg[setting])
                    widget.currentTextChanged.connect(change_setting)
                elif isinstance(widget, QPlainTextEdit):
                    widget.setPlainText(cfg[setting])
                    widget.textChanged.connect(change_plaintext_setting)
                else:
                    oslogger.warning(
                        'invalid QWidget with name {}'.format(name))
            except TypeError:
                oslogger.warning(
                    'invalid value {} for {}', format(value, setting))

    def event_setting_changed(self, setting, value):
        r"""Update the controls if settings have been changed by someone else.

        Parameters
        ----------
        setting
            The name of the changed setting
        value
            The new value
        """
        if self._its_me:
            return
        name = 'cfg_{}'.format(setting)
        if not hasattr(self.ui, name):
            return
        self._its_me = True
        widget = getattr(self.ui, name)
        try:
            if isinstance(widget, QCheckBox):
                widget.setChecked(bool(value))
            elif isinstance(widget, QLineEdit):
                widget.setText(safe_decode(value))
            elif isinstance(widget, (QSpinBox, QDoubleSpinBox)):
                widget.setValue(value)
            elif isinstance(widget, QKeySequenceEdit):
                widget.setKeySequence(QKeySequence(value))
            elif isinstance(widget, QFontComboBox):
                widget.setCurrentFont(QFont(value))
            elif isinstance(widget, QComboBox):
                widget.setCurrentText(safe_decode(value))
            elif isinstance(widget, QPlainTextEdit):
                widget.setPlainText(safe_decode(value))
            else:
                oslogger.warn('invalid QWidget with name {}'.format(name))
        except TypeError:
            oslogger.warning('invalid value {} for {}', format(value, setting))
                
        self._its_me = False



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame import misc
from libopensesame.oslogging import oslogger
from libopensesame.exceptions import InvalidValue
import tempfile
import os
import sys
import shutil


class FilePoolStore:
    """The `pool` object provides dict-like access to the file pool. When
    checking whether a file is in the file pool, several folders are
    searched.
    For more details, see `pool.folders()`.

    A `pool` object is created
    automatically when the experiment starts.

    In addition to the functions
    listed below, the following semantics are
    supported:

    __Examples__
    
    Basic use:

    ~~~ .python
    # Get the full path to a file in the file pool
    print(f'The full path to img.png is {pool["img.png"]}')
    # Check if a file is in the file pool
    if 'img.png' in pool:
        print('img.png is in the file pool')
    # Delete a file from the file pool
    del pool['img.png']
    # Walk through all files in the file pool. This retrieves the full paths.
    for path in pool:
        print(path)
    # Check the number of files in the file pool
    print(f'There are {len(pool)} files in the file pool')
    ~~~
    
    Get an image from the file pool and use a `Canvas` to show it.

    ~~~ .python
    image_path = pool['img.png']
    my_canvas = Canvas()
    my_canvas.image(image_path)
    my_canvas.show()
    ~~~
    
    [TOC]
    """
    def __init__(self, experiment, folder=None):
        r"""Constructor.

        Parameters
        ----------
        experiment : experiment
            The experiment object.
        folder : str, unicode, NoneType, optional
            The folder to use for the file pool, or `None` to create a new
            temporary folder.
        """
        self.experiment = experiment
        if folder is None:
            # On some systems tempfile.mkdtemp() triggers a UnicodeDecodeError.
            # This is resolved by passing the dir explicitly as a Unicode
            # string. This fix has been adapted from:
            # - <http://bugs.python.org/issue1681974>
            self.__folder__ = tempfile.mkdtemp(
                suffix='.opensesame_pool',
                dir=safe_decode(
                    tempfile.gettempdir(),
                    enc=sys.getfilesystemencoding()
                )
            )
            oslogger.debug('creating new pool folder')
        else:
            oslogger.debug('reusing existing pool folder')
            self.__folder__ = folder
        oslogger.debug('pool folder is \'%s\'' % self.__folder__)

    def clean_up(self):
        r"""Removes the pool folder."""
        try:
            shutil.rmtree(self.__folder__)
        except Exception:
            oslogger.error('Failed to remove %s' % self.__folder__)

    def __contains__(self, path):
        r"""Checks if a file is in the file pool.

        Returns
        -------
        bool
            A bool indicating if the file is in the pool.
        """
        if path == '':
            return False
        return os.path.exists(self[path])

    def __delitem__(self, path):
        r"""Deletes a file from the file pool.

        Parameters
        ----------
        path : str, unicode
            The name of the file.
        """
        os.remove(self[path])

    def __getitem__(self, path):
        """
        visible: False

        desc: |
            Returns the full path to a file.

        arguments:
            path:
                desc: A filename. This can be any type, but will be coerced
                      to `unicode` if it is not `unicode`.

        returns:
            desc: The full path to the file.
            type:unicode
        """
        path = safe_decode(path)
        if path.strip() == '':
            raise InvalidValue(f'Cannot get empty filename from file pool')
        for folder in self.folders(include_experiment_path=True):
            _path = os.path.normpath(os.path.join(folder, path))
            if os.path.exists(_path):
                return _path
        return os.path.normpath(path)

    def __iter__(self):
        """
        visible: False

        returns:
                type:	file_pool_store_iterator
        """
        return file_pool_store_iterator(self)

    def __len__(self):
        """
        visible: False

        returns:
                desc:	The number of files in the file pool.
                type:	int
        """
        return len(self.files())

    def count_included(self):
        """{nodoc}

        returns:
            desc: The number of files that are in the main pool folder, i.e.
                  the folder that is included with the experiment.
            type: int
        """
        return len(os.listdir(self.folder()))

    def add(self, path, new_name=None):
        r"""Copies a file to the file pool.

        Parameters
        ----------
        path : str, unicode
            The full path to the file on disk.
        new_name : str, NoneType, optional
            A new name for the file in the pool, or None to use the file's
            original name.

        Examples
        --------
        >>> pool.add('/home/username/Pictures/my_ing.png')
        """
        if new_name is None:
            new_name = os.path.basename(path)
        shutil.copyfile(path, os.path.join(self.folder(), new_name))

    def files(self):
        r"""Returns all files in the file pool.

        Returns
        -------
        list
            A list of full paths.

        Examples
        --------
        >>> for path in pool.files():
        >>>     print(path)
        >>> # Equivalent to:
        >>> for path in pool:
        >>>     print(path)
        """
        _files = []
        for _folder in self.folders():
            for _file in os.listdir(_folder):
                if os.path.isfile(os.path.join(_folder, _file)):
                    _files.append(_file)
        return sorted(_files)

    def fallback_folder(self):
        """The full path to the fallback pool folder, which is the
        `__pool__` subfolder of the current experiment folder, or
        `None` if this folder does not exist. The fallback pool
        folder is mostly useful in combination with a versioning
        system, such as git, because it allows you to save the
        experiment as a plain-text file, even when having files
        in the file pool.
        
        Examples
        --------
        >>> if pool.fallback_folder() is not None:
        >>>     print('There is a fallback pool folder!')
        
        Returns
        -------
        str
        """
        _folders = self.folders()
        if len(_folders) < 2:
            return None
        return _folders[-1]

    def folder(self):
        r"""Gives the full path to the (main) pool folder. This is typically a
        temporary folder that is deleted when the experiment is finished.

        Returns
        -------
        unicode
            The full path to the main pool folder.

        Examples
        --------
        >>> print(f'The pool folder is here: {pool.folder()}')
        """
        if not os.path.exists(self.__folder__):
            oslogger.warning('recreating missing file-pool folder')
            os.mkdir(self.__folder__)
        return self.__folder__

    def in_folder(self, path):
        r"""Checks whether path is in the pool folder. This is different from
        the `path in pool` syntax in that it only checks the main pool folder,
        and not the fallback pool folder and experiment folder.

        Parameters
        ----------
        path : str
            A file basename to check.

        Returns
        -------
        bool

        Examples
        --------
        >>> print(pool.in_folder('cue.png'))
        """
        return os.path.exists(os.path.join(self.folder(), path))

    def folders(self, include_fallback_folder=True,
                include_experiment_path=False):
        r"""Gives a list of all folders that are searched when retrieving the
        full path to a file. These are (in order):

        1. The file pool folder
        itself, as returned by `pool.folder()`.
        2. The folder of the current
        experiment (if it exists)
        3. The fallback pool folder, as returned by
        `pool.fallback_folder()` (if it exists)

        Parameters
        ----------
        include_fallback_folder : bool, optional
            Indicates whether the fallback pool folder should be included if it
            exists.
        include_experiment_path : bool, optional
            Indicates whether the experiment folder should be included if it
            exists.

        Returns
        -------
        list
            A list of all folders.

        Examples
        --------
        >>> print('The following folders are searched for files:')
        >>> for folder in pool.folders():
        >>>     print(folder)
        """
        _folders = [self.folder()]
        if self.experiment.experiment_path is None or \
                not os.path.exists(self.experiment.experiment_path):
            return _folders
        fallback_folder = os.path.join(self.experiment.experiment_path,
                                       '__pool__')
        if include_fallback_folder and os.path.exists(fallback_folder):
            _folders.append(fallback_folder)
        if include_experiment_path:
            _folders.append(self.experiment.experiment_path)
        return _folders

    def rename(self, old_path, new_path):
        r"""Renames a file in the pool folder.

        Parameters
        ----------
        old_path : str, unicode
            The old file name.
        new_path : str, unicode
            The new file name.

        Examples
        --------
        >>> pool.rename('my_old_img.png', 'my_new_img.png')
        """
        path = self[old_path]
        dirname, basename = os.path.split(path)
        os.rename(path, os.path.join(dirname, new_path))

    def size(self):
        """Gets the combined size in bytes of all files in the file pool.
        
        Examples
        --------
        >>> print(f'The size of the file pool is {pool.size()} bytes')
                
        Returns
        -------
        int
        """
        return sum([os.path.getsize(self[path]) for path in self])


class FilePoolStoreIterator:

    def __init__(self, pool):

        self.pool = pool
        self.files = pool.files()

    def __iter__(self):

        return self

    def __next__(self):

        # For Python 3
        return self.next()

    def next(self):

        if len(self.files) == 0:
            raise StopIteration
        return self.files.pop()



# Alias for backwards compatibility
file_pool_store = FilePoolStore
file_pool_store_iterator = FilePoolStoreIterator



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.sketchpad_elements._base_element import BaseElement


class Noise(BaseElement):

    r"""A gabor-patch element for the sketchpad."""
    def __init__(self, sketchpad, string):
        r"""Constructor.

        Parameters
        ----------
        sketchpad
            A sketchpad object.
        string
            A definition string.
        """
        defaults = [
            (u'x',		None),
            (u'y',		None),
            (u'env',	u'gaussian'),
            (u'size', 	96),
            (u'stdev',	12),
            (u'color1',	u'white'),
            (u'color2', u'black'),
            (u'bgmode',	u'avg')
        ]
        super().__init__(sketchpad, string, defaults=defaults)

    def draw(self):
        r"""Draws the element to the canvas of the sketchpad."""
        properties = self.eval_properties()
        return self.canvas.noise_patch(properties[u'x'], properties[u'y'],
                                       env=properties[u'env'],
                                       size=properties[u'size'],
                                       stdev=properties[u'stdev'],
                                       col1=properties[u'color1'],
                                       col2=properties[u'color2'],
                                       bgmode=properties[u'bgmode'])


# Alias for backwards compatibility
noise = Noise



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import os
from libqtopensesame.misc.config import cfg
from libqtopensesame.extensions import BaseExtension
from libopensesame.oslogging import oslogger
from qtpy.QtWidgets import QDockWidget, QShortcut
from qtpy.QtCore import Qt
from qtpy.QtGui import QKeySequence
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'JupyterConsole', category=u'extension')


class JupyterConsole(BaseExtension):
    
    preferences_ui = 'extensions.jupyter_console.preferences'

    @BaseExtension.as_thread(wait=500)
    def event_startup(self):

        from .jupyter_tabwidget import ConsoleTabWidget

        self.set_busy(True)
        self._jupyter_console = ConsoleTabWidget(self.main_window)
        self._dock_widget = QDockWidget(u'Console', self.main_window)
        self._dock_widget.setObjectName(u'JupyterConsole')
        self._dock_widget.setWidget(self._jupyter_console)
        self._dock_widget.closeEvent = self._on_close_event
        self.main_window.addDockWidget(
            Qt.BottomDockWidgetArea,
            self._dock_widget
        )
        self._set_visible(cfg.jupyter_visible)
        self._shortcut_focus = QShortcut(
            QKeySequence(cfg.jupyter_focus_shortcut),
            self.main_window,
            self._focus,
            context=Qt.ApplicationShortcut
        )
        self.set_busy(False)
        
    def fire(self, event, **kwdict):
        
        if event != 'startup':
            oslogger.debug('ignoring events until after startup')
            return
        JupyterConsole.fire = BaseExtension.fire
        self.fire(event, **kwdict)

    def activate(self):

        if not hasattr(self, '_jupyter_console'):
            oslogger.debug('ignoring activate until after startup')
            return
        self._set_visible(not cfg.jupyter_visible)

    def event_run_experiment(self, fullscreen):

        oslogger.debug(u'capturing stdout')
        self._jupyter_console.current.capture_stdout()

    def event_end_experiment(self, ret_val):

        self._jupyter_console.current.release_stdout()
        self._jupyter_console.current.show_prompt()
        oslogger.debug(u'releasing stdout')

    def event_jupyter_start_kernel(self, kernel):

        self._jupyter_console.add(kernel=kernel)

    def event_jupyter_run_file(self, path, debug=False):

        self._set_visible(True)
        if not os.path.isfile(path):
            return
        self._jupyter_console.current.change_dir(os.path.dirname(path))
        if debug:
            self._jupyter_console.current.run_debug(
                path,
                breakpoints=self.extension_manager.provide(
                    'pyqode_breakpoints'
                )
            )
        else:
            self._jupyter_console.current.run_file(path)

    def event_jupyter_change_dir(self, path):

        self._jupyter_console.current.change_dir(path)

    def event_jupyter_run_code(self, code, editor=None):

        self._set_visible(True)
        self._jupyter_console.current.execute(code)
    
    def event_jupyter_run_silent(self, code):

        self._jupyter_console.current.execute(code)
        
    def event_jupyter_run_system_command(self, cmd):

        self._jupyter_console.current.run_system_command(cmd)

    def event_jupyter_write(self, msg):

        try:
            self._jupyter_console.current.write(msg)
        except AttributeError:
            oslogger.error(safe_decode(msg))

    def event_jupyter_focus(self):

        self._jupyter_console.current.focus()

    def event_jupyter_show_prompt(self):

        self._jupyter_console.current.show_prompt()

    def event_jupyter_restart(self):

        self._jupyter_console.current.restart()

    def event_jupyter_interrupt(self):

        self._jupyter_console.current.interrupt()

    def event_set_workspace_globals(self, global_dict):

        self._jupyter_console.current.set_workspace_globals(global_dict)

    def provide_jupyter_workspace_name(self):

        try:
            return self._jupyter_console.current.name
        except AttributeError:
            return None
        
    def provide_workspace_kernel(self):
        
        try:
            return self._jupyter_console.current._kernel
        except AttributeError:
            return None
        
    def provide_workspace_language(self):

        try:
            return self._jupyter_console.current.language
        except AttributeError:
            return None
        
    def provide_workspace_logging_commands(self):

        from .jupyter_tabwidget.constants import LOGGING_LEVEL_CMD
        try:
            kernel = self._jupyter_console.current.language
        except AttributeError:
            return None
        return LOGGING_LEVEL_CMD.get(kernel, None)

    def provide_jupyter_workspace_globals(self):

        return self.get_workspace_globals()

    def provide_jupyter_list_workspace_globals(self):

        return self.list_workspace_globals()

    def provide_jupyter_workspace_variable(self, name):

        return self._jupyter_console.current.get_workspace_variable(name)
        
    def provide_jupyter_kernel_running(self):
        
        try:
            return self._jupyter_console.current.running()
        except AttributeError:
            return False

    def provide_jupyter_check_syntax(self, code):

        return self._jupyter_console.current.check_syntax(code)

    def get_workspace_globals(self):

        try:
            return self._jupyter_console.current.get_workspace_globals()
        except AttributeError:
            return {u'no reply': None}

    def list_workspace_globals(self):

        try:
            return self._jupyter_console.current.list_workspace_globals()
        except Exception as e:
            print(e)
            return []

    def event_close(self):

        if not hasattr(self, '_jupyter_console'):
            oslogger.debug('ignoring close all')
            return
        self._jupyter_console.close_all()

    def _set_visible(self, visible):

        cfg.jupyter_visible = visible
        self.set_checked(visible)
        if visible:
            self._dock_widget.show()
            self._jupyter_console.current.focus()
        else:
            self._dock_widget.hide()

    def _focus(self):

        self._set_visible(True)
        self._jupyter_console.current.focus()

    def _on_close_event(self, e):

        self._set_visible(False)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtGui
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent


class BaseValidator(BaseSubcomponent, QtGui.QValidator):

    r"""A base class for input validators."""
    def __init__(self, main_window, default):

        self.default = default
        main_window = self.get_main_window(main_window)
        super().__init__(main_window)
        self.setup(main_window)

    def is_valid(self, val):

        return True

    def validate(self, val, pos):

        if self.is_valid(val):
            return self.Acceptable, val, pos
        return self.Intermediate, val, pos

    def fixup(self, val):

        if not self.is_valid(val):
            return self.default
        return val



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.dialogs.base_dialog import BaseDialog


class PatchSettings(BaseDialog):

    r"""A patch settings dialog, inherited by the gabor and noise settings."""
    # Maps the GUI names onto valid keyword values
    env_map = {
        u'gaussian': u'gaussian',
        u'linear': u'linear',
        u'circular (sharp edge)': u'circular',
        u'rectangular (no envelope)': u'rectangular',
    }
    bgmode_map = {
        u'Color average': u'avg',
        u'Color 2': u'col2',
    }

    def __init__(self, main_window):
        r"""Constructor.

        Parameters
        ----------
        main_window
            The main window object.
        """
        super().__init__(main_window, ui=self.ui_file())
        self.ui.combobox_env.currentIndexChanged.connect(self.apply_env)
        self.ui.edit_color1.initialize()
        self.ui.edit_color2.initialize(color=self.experiment.var.background)
        self.ui.apply_env()

    def ui_file(self):
        """
        returns:
                desc:	The ui file.
                type:	unicode
        """
        raise NotImplementedError()

    def apply_env(self):
        r"""Enables/ disables the stdev control based on the env combobox."""
        self.ui.spinbox_stdev.setEnabled(
            self.ui.combobox_env.currentText() == u'gaussian')

    def get_properties(self):
        r"""Gets the Gabor properties.

        Returns
        -------
        dict
            A dictionary with properties.
        """
        raise NotImplementedError()

    def set_properties(self, properties):
        r"""Fills the dialog controls based on a properties dictionary.

        Parameters
        ----------
        properties : dict
            A properties dictionary.
        """
        raise NotImplementedError()

    def _spinbox_properties(self, *spinbox_vars):
        r"""Gets the properties of all enabled spinboxes.

        Parameters
        ----------
        *spinbox_vars : list
            A list of variables that are defined through spinbox widgets.

        Returns
        -------
        A dictionary with (var, value) mappings
        """
        return {
            var: getattr(self.ui, u'spinbox_%s' % var).value()
            for var in spinbox_vars
            if getattr(self.ui, u'spinbox_%s' % var).isEnabled()
        }

    def _combobox_properties(self, *combobox_vars):
        r"""Gets the properties of all enabled comboboxes.

        Parameters
        ----------
        *combobox_vars : list
            A list of (var, options) tuples of variables that are defined
            through combobox widgets.

        Returns
        -------
        A dictionary with (var, value) mappings
        """
        return {
            var: options[getattr(self.ui, u'combobox_%s' % var).currentText()]
            for var, options in combobox_vars
            if getattr(self.ui, u'combobox_%s' % var).isEnabled()
        }

    def _set_spinbox(self, spinbox, var, properties):
        r"""Safely sets a spinbox based on a variable, or disables the spinbox
        if the value is not numeric.

        Parameters
        ----------
        spinbox
            A QSpinBox
        var
            A variable name
        properties
            A properties dict
        """
        if isinstance(properties[var], (int, float)):
            spinbox.setEnabled(True)
            spinbox.setValue(properties[var])
        else:
            spinbox.setEnabled(False)

    def _set_combobox(self, options, combobox, var, properties):
        r"""Safely sets a combobox based on a variable, or disables the
        combobox if the value is not among the combobox options.

        Parameters
        ----------
        options
            A dictionary with options
        combobox
            QComboBox
        var
            A variable name
        properties
            A properties dict
        """
        for key, val in options.items():
            if val == properties[var]:
                i = combobox.findText(key)
                combobox.setCurrentIndex(i)
                combobox.setEnabled(True)
                return
        combobox.setEnabled(False)


# Alias for backwards compatibility
patch_settings = PatchSettings



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'quick_switcher', category=u'extension')


class QuickSwitcher(BaseExtension):

    r"""The quick-switcher allows you to quickly navigate to items."""
    def activate(self):

        haystack = []
        for item in self.experiment.items.values():
            haystack.append((
                u'{} ({})'.format(item.name, item.item_type),
                item,
                item.open_tab
            ))
        self.extension_manager.fire(
            u'quick_select',
            haystack=haystack,
            placeholder_text=_(u'Search items …')
        )



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
import sys
import platform
import multiprocessing

if platform.system() == 'Linux':
    # The fork multiprocessing method, which is the default on Linux, seems
    # unstable on some systems. Therefore, we use spawn as the default.
    # However, setting the OPENSESAME_MULTIPROCESSING_METHOD environment
    # variable allows users to customize this. See also:
    # - <https://github.com/open-cogsci/OpenSesame/issues/782>
    try:
        multiprocessing.set_start_method(
            os.environ.get('OPENSESAME_MULTIPROCESSING_METHOD', 'spawn'))
    except RuntimeError as e:
        print(f'failed to change multiprocessing start method: {e}')
    except ValueError as e:
        print(f'invalid multiprocessing start method: {e}')
    # solves a library conflict for Linux with Nvidia drivers
    # See https://forum.qt.io/topic/81328/ubuntu-qopenglshaderprogram-shader-program-is-not-linked/2
    try:
        from OpenGL import GL
    except ImportError:  # Rapunzel doesn't need OpenGL
        pass

# Attach a dummy console when launched with pythonw.exe. This is necessary for
# some libraries.
if sys.executable.endswith('pythonw.exe'):
    sys.stdout = open(os.devnull, 'w')
    sys.stderr = open(os.devnull, 'w')
    sys.stdin = open(os.devnull)

# On MacOS, the locations of the SSL cerfificates need to be explicitly
# indicated when packaged # as an app, otherwise urllib.urlopen fails with an
# SSL related error. To do so, we make use of the certifi package that can
# provide this location, but it is unclear if certifi is always installed,
# so let's fail gracefully if it isn't.
if platform.system() == 'Darwin':
    try:
        import certifi
        os.environ['SSL_CERT_FILE'] = certifi.where()
        os.environ['SSL_CERT_DIR'] = os.path.dirname(certifi.where())
    except ImportError:
        pass  # Or log something here?
    # Workaround for an incompatibility between Qt and Mac OS Big Sur. See:
    # - https://forum.cogsci.nl/discussion/comment/21525/#Comment_21525
    os.environ['QT_MAC_WANTS_LAYER'] = '1'

from openexp import resources
from pathlib import Path
# Add the libqtopensesame resource folder, which contains ui files etc
resources.add_resource_folder(Path(__file__).parent / 'resources')

def patch_pyqt():
    """This patches PyQt such that properties are removed from objects before
    connectSlotsByName() tries to inspect the object. This is necessary because
    in some versions, the introspection crashes when properties cannot be
    accessed, for example because the object needs to be initialized first.
    """
    def _(fnc):

        def make_object_safe(obj, cls):

            tmp = {}
            for name, value in list(cls.__dict__.items()):
                try:
                    getattr(obj, name)
                except AttributeError:
                    tmp[name] = value
                    delattr(cls, name)
            return tmp

        def restore_object(cls, tmp):

            for name, value in tmp.items():
                setattr(cls, name, value)

        def inner(obj):

            tmp = {}
            for cls in obj.__class__.mro():
                tmp[cls] = make_object_safe(obj, cls)
            fnc(obj)
            for cls in tmp:
                restore_object(cls, tmp[cls])

        return inner

    from qtpy import QtCore
    QtCore.QMetaObject.connectSlotsByName = \
        _(QtCore.QMetaObject.connectSlotsByName)


def set_paths():

    from libopensesame import misc
    import sys
    import platform
    # On Windows, the working directory should be the folder with the Python
    # interpreter. If not, Qt will have difficulty finding the correct paths.
    # We can disable this behavior on the command line, because it will
    # actually break things on some installations.
    if platform.system() == 'Windows' and '--no-chdir' not in sys.argv:
        os.chdir(os.path.dirname(sys.executable))
    # Update the system path so that Qt can find itself
    path = os.path.join(os.getcwd(), 'Library', 'bin')
    if os.path.exists(path):
        os.environ['PATH'] = path + ';' + os.environ['PATH']
    # Add the Scripts subfolder, which is where Anaconda scripts are located
    path = os.path.join(os.getcwd(), 'Scripts')
    if os.path.exists(path):
        os.environ['PATH'] = path + ';' + os.environ['PATH']
    from qtpy import QtCore
    # Add the folder that contains the OpenSesame modules to the path. This is
    # generally only necessary if OpenSesame is directly run from source,
    # instead from an installation.
    if os.path.exists(os.path.join(os.getcwd(), 'libopensesame')):
        sys.path.insert(0, os.getcwd())
    # Add the Qt plugin folders to the library path, if they exists. Where
    # these folders are depends on the version of Qt4, but these are two
    # possible locations.
    qt_plugin_path = os.path.join(
        os.path.dirname(sys.executable), 'Library', 'plugins')
    if os.path.isdir(qt_plugin_path):
        QtCore.QCoreApplication.addLibraryPath(
            safe_decode(qt_plugin_path, enc=sys.getfilesystemencoding()))
    qt_plugin_path = os.path.join(
        os.path.dirname(sys.executable), 'Library', 'lib', 'qt4', 'plugins')
    if os.path.isdir(qt_plugin_path):
        QtCore.QCoreApplication.addLibraryPath(
            safe_decode(qt_plugin_path, enc=sys.getfilesystemencoding()))


def opensesame():
    """The entrty point for the OpenSesame GUI"""
    set_paths()
    patch_pyqt()
    # Support for multiprocessing when packaged
    # In OS X the multiprocessing module is horribly broken, but a fixed
    # version has been released as the 'billiard' module
    if platform.system() == 'Darwin':
        # Use normal multiprocessing module from python 3.4 and on
        if sys.version_info >= (3, 4):
            from multiprocessing import freeze_support, set_start_method
            freeze_support()
            set_start_method('spawn')
        else:
            from billiard import freeze_support, forking_enable
            freeze_support()
            forking_enable(0)
    else:
        from multiprocessing import freeze_support
        freeze_support()
    # Parse the (optional) environment file that contains special paths, etc.
    from libopensesame import misc
    misc.parse_environment_file()
    # Do the basic window initialization
    from qtpy.QtWidgets import QApplication
    from qtpy.QtCore import Qt
    # From Qt 5.6 on, QtWebEngine is the default way to render web pages
    # QtWebEngineWidgets must be imported before a QCoreApplication instance is
    # created.
    try:
        from qtpy import QtWebEngineWidgets
    except ImportError:
        pass
    # Enable High DPI pixmaps with PyQt5
    if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps)
    # Enable scaling for High DPI displays with PyQt5
    if hasattr(Qt, 'AA_EnableHighDpiScaling'):
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
    # Enable proportional scaling so that things don't look disproportionately
    # large on Windows
    if hasattr(Qt, 'HighDpiScaleFactorRoundingPolicy'):
        QApplication.setHighDpiScaleFactorRoundingPolicy(
            Qt.HighDpiScaleFactorRoundingPolicy.PassThrough)
    from libqtopensesame.qtopensesame import QtOpenSesame
    app = QApplication(sys.argv)
    opensesame = QtOpenSesame(app)
    opensesame.__script__ = __file__
    app.processEvents()
    # Install the translator. For some reason, the translators need to be
    # instantiated here and not in the set_locale() function, otherwise the
    # locale is ignored.
    from qtpy.QtCore import QTranslator
    opensesame.set_locale([QTranslator(), QTranslator(), QTranslator(),
                           QTranslator(), QTranslator(), QTranslator(),
                           QTranslator(), QTranslator(), QTranslator(),
                           QTranslator(), QTranslator(), QTranslator()])
    # Now that the window is shown, load the remaining modules and resume the
    # GUI initialization.
    opensesame.resume_init()
    opensesame.restore_window_state()
    opensesame.show()
    # Added for OS X, otherwise Window will not appear
    opensesame.raise_()
    # Exit using the application exit status
    sys.exit(app.exec_())


def opensesamerun():
    """The entrty point for the OpenSesameRun GUI"""
    set_paths()
    from libopensesame.oslogging import oslogger
    oslogger.start('gui')
    from libopensesame import misc
    misc.parse_environment_file()
    from libopensesame.experiment import Experiment
    # Parse the command line options
    options = misc.opensesamerun_options()
    app = None
    # If the command line options haven't provided sufficient information to
    # run right away, present a GUI
    while not misc.opensesamerun_ready(options):
        # If PyQt is not available (e.g., this might be the case on Mac OS)
        # give an error instead of showing a GUI. This makes sure that even
        # without PyQt, people can still run experiments.
        try:
            from qtpy import QtGui, QtCore, QtWidgets
        except Exception as e:
            misc.messagebox("OpenSesame Run", "Incorrect or "
                "missing options.\n\nRun 'opensesame --help' from a terminal "
                "(or command prompt) to see a list of available options, or "
                "install PyQt to enable the graphical user interface.")
            sys.exit()
        # Create the GUI and show it
        from libqtopensesame.qtopensesamerun import QtOpenSesameRun
        if app is None:
            app = QtWidgets.QApplication(sys.argv)
            myapp = QtOpenSesameRun(options)
        myapp.show()
        app.exec_()
        # Update the options from the GUI
        options = myapp.options
        # Exit if the GUI was canceled
        if not myapp.run:
            sys.exit()
    # Decode the experiment path and logfile
    experiment = os.path.abspath(options.experiment)
    if isinstance(experiment, bytes):
        experiment = safe_decode(experiment, enc=sys.getfilesystemencoding(),
                                 errors='ignore')
    # experiment_path = os.path.dirname(experiment)
    logfile = options.logfile
    if isinstance(logfile, bytes):
        logfile = safe_decode(logfile, enc=sys.getfilesystemencoding(),
                              errors='ignore')
    experiment_path = safe_decode(os.path.abspath(options.experiment),
                                  enc=sys.getfilesystemencoding())
    # In debug mode, don't try to catch any exceptions
    if options.debug:
        exp = Experiment("Experiment", experiment,
                         experiment_path=experiment_path)
        exp.set_subject(options.subject)
        exp.var.fullscreen = options.fullscreen
        exp.logfile = logfile
        exp.run()
        exp.end()
        return
    # Try to parse the experiment from a file
    try:
        exp = Experiment("Experiment", experiment,
                         experiment_path=experiment_path)
    except Exception as e:
        misc.messagebox(
            "OpenSesame Run",
            misc.strip_tags(e)
        )
        sys.exit()
    # Set some options
    exp.set_subject(options.subject)
    exp.var.fullscreen = 'yes' if options.fullscreen else 'no'
    exp.logfile = logfile
    # Initialize random number generator
    import random
    random.seed()
    # Try to run the experiment
    try:
        exp.run()
    except Exception as e:
        misc.messagebox("OpenSesame Run", misc.strip_tags(e))
    finally:
        try:
            exp.end()
        except Exception as f:
            misc.messagebox("OpenSesame Run", misc.strip_tags(f))
    exp.pool.clean_up()



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent


class ConsoleBridge(BaseSubcomponent):

    r"""A base console for debug-window consoles."""
    def __init__(self, main_window):

        super(ConsoleBridge, self).__init__()
        self.setup(main_window)
        self._jupyter_console = None
        self._writing = False

    def write(self, s):

        oslogger.debug(s)
        if self._writing:
            oslogger.warning(u'recursive write() call')
            print(safe_decode(s, errors=u'ignore'))
            return
        self._writing = True
        if hasattr(self, u'extension_manager'):
            self.extension_manager.fire(u'jupyter_write', msg=s)
        else:
            oslogger.info(s)
        self._writing = False

    def reset(self):

        self.main_window.set_run_status(u'inactive')
        self.extension_manager.fire(u'jupyter_restart')

    def get_workspace_globals(self):

        if self._jupyter_console is None:
            try:
                self._jupyter_console = self.extension_manager['JupyterConsole']
            except Exception:
                return {}
        return self._jupyter_console.get_workspace_globals()

    def set_workspace_globals(self, _globals={}):

        self.extension_manager.fire(
            u'set_workspace_globals',
            global_dict=_globals
        )

    def show_prompt(self):

        self.extension_manager.fire(u'jupyter_show_prompt')



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from ..form_base.form_base import FormBase

default_script = u"""
set form_question 'Your question'
set form_title Title
set form_var response
set rows 1;1;6
set cols 1
widget 0 0 1 1 label text={form_title}
widget 0 1 1 1 label text={form_question} center=no
widget 0 2 1 1 text_input return_accepts=yes focus=yes var={form_var} stub=""
"""


class FormTextInput(FormBase):

    initial_view = u'controls'

    def __init__(self, name, experiment, string=None):
        if string is None or string.strip() == u'':
            string = default_script
        super().__init__(name, experiment, string)

    def var_info(self):
        return super().var_info() + \
            [(self.var.get(u'form_var', _eval=False, default=u''),
             u'[Depends on response]')]



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtGui, QtCore, QtWidgets
from openexp import resources
from libqtopensesame.misc.base_component import BaseComponent
from libqtopensesame.misc.config import cfg
from libqtopensesame.items.experiment import Experiment
from libopensesame import metadata
from libopensesame.exceptions import OSException
from libopensesame.oslogging import oslogger
from libopensesame import misc
import os
import sys
import warnings
import platform
import traceback
import subprocess
from libqtopensesame.misc.translate import translation_context
_ = translation_context('qtopensesame', category='core')
oslogger.start('gui')


SAVE_FILE_FILTER = 'OpenSesame files (*.osexp)'
OPEN_FILE_FILTER = (
    'OpenSesame files (*.osexp *.opensesame.tar.gz *.opensesame);;'
    'Python scripts (*.py);;'
    'R scripts (*.R);;'
    'All files (*.*)')


class QtOpenSesame(QtWidgets.QMainWindow, BaseComponent):
    """The main class of the OpenSesame GUI.
    
    Parameters
    ----------
    app : QApplication
    parent : QObject, optional
    """
    
    def __init__(self, app, parent=None):
        """Constructor. This does very little, except prepare the app to be 
        shown as rapidly as possible. The actual GUI initialization is handled
        by resume_init().
        """
        if sys.platform == 'darwin':
            # Workaround for Qt issue on OS X that causes QMainWindow to
            # hide when adding QToolBar, see
            # https://bugreports.qt-project.org/browse/QTBUG-4300
            QtWidgets.QMainWindow.__init__(
                self,
                parent,
                QtCore.Qt.MacWindowToolBarButtonHint
            )
        else:
            QtWidgets.QMainWindow.__init__(self, parent)
        self._locale = None
        self.app = app
        self.first_show = True
        self.current_path = None
        self.version = metadata.__version__
        self.codename = metadata.codename
        self.lock_refresh = False
        self.unsaved_changes = False
        self._run_status = 'inactive'
        self.block_close_event = False
        self.parse_command_line()
        self.restore_config()

    def resume_init(self):
        """Resume GUI initialization"""
        import opensesame_extensions
        from libopensesame import misc
        from libopensesame.plugin_manager import PluginManager
        from libqtopensesame.misc import theme
        from libqtopensesame.misc.console_bridge import ConsoleBridge
        from libqtopensesame.widgets.pool_widget import PoolWidget
        from libqtopensesame.extensions import ExtensionManager
        import random

        # Make sure that icons are shown in context menu, regardless of the
        # system settings. This is necessary, because Ubuntu doesn't show menu
        # icons by default.
        QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_DontShowIconsInMenus,
                                            False)
        # Initialize random number generator
        random.seed()
        # Check the filesystem encoding for debugging purposes
        oslogger.debug(f'filesystem encoding: {sys.getfilesystemencoding()}')
        # self.restore_config()
        self.set_style()
        self.set_warnings()
        # Setup the UI
        self.load_ui('misc.main_window')
        self.theme = theme.theme(self, self.options._theme)
        self.ui.itemtree.setup(self)
        self.ui.tabwidget.main_window = self
        # Determine the home folder
        self.home_folder = misc.home_folder()
        # Create .opensesame folder if it doesn't exist yet
        if not os.path.exists(os.path.join(self.home_folder, ".opensesame")):
            os.mkdir(os.path.join(self.home_folder, ".opensesame"))
        # Set the window message
        self._read_only = False
        self.window_message(_("New experiment"))
        # Set the window icon
        self.setWindowIcon(self.theme.qicon("opensesame"))
        # Make the connections
        self.ui.itemtree.structure_change.connect(self.update_overview_area)
        self.ui.action_quit.triggered.connect(self.close)
        self.ui.action_open.triggered.connect(self.open_file)
        self.ui.action_save.triggered.connect(self.save_file)
        self.ui.action_save_as.triggered.connect(self.save_file_as)
        self.ui.action_new_window.triggered.connect(self.new_window)
        self.ui.action_run.triggered.connect(self.run_experiment)
        self.ui.action_run_in_window.triggered.connect(
            self.run_experiment_in_window)
        self.ui.action_run_quick.triggered.connect(self.run_quick)
        self.ui.action_kill.triggered.connect(self.kill_experiment)
        self.ui.action_close_current_tab.triggered.connect(
            self.ui.tabwidget.close_current)
        self.ui.action_close_all_tabs.triggered.connect(
            self.ui.tabwidget.close_all)
        self.ui.action_close_other_tabs.triggered.connect(
            self.ui.tabwidget.close_other)
        self.ui.action_onetabmode.triggered.connect(
            self.ui.tabwidget.toggle_onetabmode)
        self.ui.action_show_overview.triggered.connect(self.toggle_overview)
        self.ui.action_show_pool.triggered.connect(
            self.toggle_pool)
        self.ui.action_preferences.triggered.connect(
            self.ui.tabwidget.open_preferences)
        # Setup the overview area
        self.ui.dock_overview.show()
        self.ui.dock_overview.visibilityChanged.connect(
            self.ui.action_show_overview.setChecked)
        # Setup the file pool
        self.ui.dock_pool.hide()
        self.ui.dock_pool.visibilityChanged.connect(
            self.ui.action_show_pool.setChecked)
        self.ui.pool_widget = PoolWidget(self)
        self.ui.dock_pool.setWidget(self.ui.pool_widget)
        # Initialize keyboard shortcuts
        self.ui.shortcut_itemtree = self._kb_shortcut(self.focus_overview_area)
        self.ui.shortcut_tabwidget = self._kb_shortcut(self.ui.tabwidget.focus)
        self.ui.shortcut_pool = self._kb_shortcut(self.focus_file_pool)
        # Add shortcuts to tooltips
        self._tooltip_shortcut(self.action_open)
        self._tooltip_shortcut(self.action_save)
        self._tooltip_shortcut(self.action_save_as)
        self._tooltip_shortcut(self.action_run)
        self._tooltip_shortcut(self.action_run_in_window)
        self._tooltip_shortcut(self.action_run_quick)
        self._tooltip_shortcut(self.action_close_other_tabs)
        self._tooltip_shortcut(self.action_onetabmode)
        self._tooltip_shortcut(self.action_show_pool)
        self._tooltip_shortcut(self.action_show_overview)
        # Create the initial experiment, which is the default template.
        with safe_open(resources['templates/default.osexp'], 'r') as fd:
            self.experiment = Experiment(self, 'New experiment', fd.read())
        self.experiment.build_item_tree()
        self.ui.itemtree.default_fold_state()
        # Miscellaneous initialization
        self.restore_state()
        self.update_recent_files()
        self.set_unsaved(False)
        self.init_custom_fonts()
        self.console = ConsoleBridge(self)
        self._unloaded_extension_manager = PluginManager(opensesame_extensions)
        self.extension_manager = ExtensionManager(self)
        self.extension_manager.register_extension(self.ui.toolbar_items)
        self.extension_manager.fire('startup')

    @property
    def mode(self):
        """A property that determines the application mode, falling back to
        'default'.
        """
        if self.options.mode is None:
            return 'default'
        return self.options.mode

    def _tooltip_shortcut(self, action):
        """Adds a shortcut between parentheses to the tooltip of an action.

        Parameters
        ----------
        action
            A QAction
        """
        if not action.shortcut() or action.toolTip()[-1] == ')':
            return
        action.setToolTip(
            '{} ({})'.format(action.toolTip(), action.shortcut().toString())
        )

    def _kb_shortcut(self, target):
        """Builds a simple keyboard shortcut. The key sequence is specified
        later, while restoring the settings.

        Parameters
        ----------
        target
            The target function.
        """
        return QtWidgets.QShortcut(QtGui.QKeySequence(), self, target)

    def focus_overview_area(self):
        """Shows and focuses the overview area."""
        self.ui.itemtree.setFocus()
        self.ui.dock_overview.setVisible(True)

    def focus_file_pool(self):
        """Shows and focuses the file pool."""
        self.ui.pool_widget.setFocus()
        self.ui.dock_pool.setVisible(True)

    def init_custom_fonts(self):
        """Registers the custom OpenSesame fonts, so that they are properly
        displayed in the sketchpad widget.
        """
        from libqtopensesame.widgets.font_widget import font_widget
        # The last element of font_list is the "other" entry
        for font in font_widget.font_list[:-1] + [
                'RobotoCondensed-Regular',
                'RobotoSlab-Regular',
                'RobotoMono-Regular',
                'Roboto-Regular'
        ]:
            try:
                ttf = resources[f'{font}.ttf']
            except FileNotFoundError:
                oslogger.error('failed to find %s' % font)
            else:
                oslogger.debug('registering %s (%s)' % (font, ttf))
                id = QtGui.QFontDatabase.addApplicationFont(ttf)
                families = QtGui.QFontDatabase.applicationFontFamilies(id)
                if families:
                    QtGui.QFont.insertSubstitution(font, families[0])

    def parse_command_line(self):
        """Parse command line options"""
        import optparse

        parser = optparse.OptionParser(
            "usage: opensesame [experiment] [options]",
            version="%s '%s'" % (self.version, self.codename))
        parser.set_defaults(debug=False)
        group = optparse.OptionGroup(parser, "Miscellaneous options")
        group.add_option(
            "-c",
            "--config",
            action="store",
            dest="_config",
            help=(
                "Set a configuration option, e.g, '--config auto_update_check="
                "False;scintilla_font_size=10'. For a complete list of "
                "configuration options, please refer to the source of config.py."
            )
        )
        group.add_option(
            "-p",
            "--profile",
            action="store",
            dest="_config_profile",
            help=("Specify a profile with its own configuration")
        )
        group.add_option(
            "-t",
            "--theme",
            action="store",
            dest="_theme",
            help="Specify a GUI theme"
        )
        group.add_option(
            "-d",
            "--debug",
            action="store_true",
            dest="debug",
            help="Print lots of debugging messages to the standard output"
        )
        group.add_option(
            "--start-clean",
            action="store_true",
            dest="start_clean",
            help="Do not load configuration and do not restore window geometry"
        )
        group.add_option(
            "--locale",
            action="store_true",
            dest="locale",
            help="Specify localization"
        )
        group.add_option(
            "--catch-translatables",
            action="store_true",
            dest="catch_translatables",
            help="Log all translatable text (for developers)"
        )
        group.add_option(
            "--performance-profile",
            action="store_true",
            dest="profile",
            help="Profile OpenSesame performance (for developers)"
        )
        group.add_option(
            "--no-global-resources",
            action="store_true",
            dest="no_global_resources",
            help="Do not use global resources on *nix"
        )
        group.add_option(
            '-w',
            "--warnings",
            action="store_true",
            dest="warnings",
            help="Show elaborate warnings"
        )
        group.add_option(
            "--mode",
            action="store",
            dest="mode",
            help="Specify the application mode (default or ide)"
        )
        group.add_option(
            "--no-chdir",
            action="store_true",
            default="false",
            dest="no_chdir",
            help="Don't change the working directory to that of the Pythone executable (Windows only)'"
        )
        parser.add_option_group(group)
        self.options, args = parser.parse_args(sys.argv)

    def set_warnings(self):
        """Sets a custom warning function, if specified on the command line."""
        if '-w' not in sys.argv and '--warnings' not in sys.argv:
            return
        warnings.simplefilter('always')
        import traceback

        def warn_with_traceback(message, category, filename, lineno, line=None):
            print('***startwarning***')
            traceback.print_stack()
            print(
                warnings.formatwarning(
                    message,
                    category,
                    filename,
                    lineno,
                    line
                )
            )
            print('***endwarning***')

        warnings.showwarning = warn_with_traceback

    def restore_config(self):
        """Restores the configuration settings, but doesn't apply anything"""
        if self.options.start_clean:
            return
        cfg.restore(
            self.mode
            if self.options._config_profile is None
            else self.options._config_profile
        )

    def restore_state(self):
        """Restore the current window to the saved state"""
        # Force configuration options that were set via the command line
        cfg.parse_cmdline_args(self.options._config)
        self.recent_files = []
        if self.options.start_clean:
            oslogger.info('Not restoring state')
            self.theme.set_toolbar_size(cfg.toolbar_size)
            return
        self.resize(cfg.size)
        self.move(cfg.pos)
        # Set the keyboard shortcuts
        self.ui.shortcut_itemtree.setKey(QtGui.QKeySequence(
            cfg.shortcut_itemtree))
        self.ui.shortcut_tabwidget.setKey(QtGui.QKeySequence(
            cfg.shortcut_tabwidget))
        self.ui.shortcut_pool.setKey(QtGui.QKeySequence(cfg.shortcut_pool))

        # Unpack the string with recent files and only remember those that exist
        recent_files = cfg.recent_files
        if hasattr(recent_files, "split"):
            for path in recent_files.split(";;"):
                if os.path.exists(path):
                    oslogger.debug("adding recent file '%s'" % path)
                    self.recent_files.append(path)
                else:
                    oslogger.debug("missing recent file '%s'" % path)
        # On Mac OS we always enable one-tab mode to deal with an issue in
        # Qt that makes the tabbar grow too large.
        if platform.system() == 'Darwin':
            cfg.onetabmode = True
            self.ui.action_onetabmode.setVisible(False)
        self.ui.action_onetabmode.setChecked(cfg.onetabmode)
        self.ui.tabwidget.toggle_onetabmode()
        if cfg.toolbar_text:
            self.ui.toolbar_main.setToolButtonStyle(
                QtCore.Qt.ToolButtonTextUnderIcon)
        else:
            self.ui.toolbar_main.setToolButtonStyle(
                QtCore.Qt.ToolButtonIconOnly)
        self.theme.set_toolbar_size(cfg.toolbar_size)

    def restore_window_state(self):
        """
        This is done separately from the rest of the restoration, because if we
        don't wait until the end, the window gets distorted again.
        """
        if self.options.start_clean:
            oslogger.info('Not restoring window state')
            return
        self.restoreState(cfg._initial_window_state)
        self.restoreGeometry(cfg._initial_window_geometry)

    def save_state(self):
        """Restores the state of the current window"""
        cfg.size = self.size()
        cfg.pos = self.pos()
        cfg._initial_window_geometry = self.saveGeometry()
        cfg._initial_window_state = self.saveState()
        cfg.toolbar_text = self.ui.toolbar_main.toolButtonStyle() == \
            QtCore.Qt.ToolButtonTextUnderIcon
        cfg.recent_files = ";;".join(self.recent_files)
        cfg.save()

    def is_busy(self):
        """
        returns:
                True if the cursor has a busy state, False otherwise
        """
        return QtWidgets.QApplication.overrideCursor() is not None

    def set_busy(self, state=True):
        """
        Show/ hide the busy notification

        Keywords arguments:
        state -- indicates the busy status (default=True)
        """
        if state:
            QtWidgets.QApplication.setOverrideCursor(QtGui.QCursor(
                QtCore.Qt.WaitCursor))
        else:
            QtWidgets.QApplication.restoreOverrideCursor()
        QtWidgets.QApplication.processEvents()

    def set_style(self):
        """Apply the application style"""
        if cfg.style in QtWidgets.QStyleFactory.keys():
            self.setStyle(QtWidgets.QStyleFactory.create(cfg.style))
            oslogger.debug("using style '%s'" % cfg.style)
        else:
            cfg.style = ''

    @property
    def locale(self):

        if self._locale:
            return self._locale
        # If a locale has been explicitly specified, use it; otherwise, use the
        # system default locale.
        locale = (cfg.locale if cfg.locale else QtCore.QLocale().bcp47Name())
        # If a locale has been specified on the command line, it overrides.
        for i, argv in enumerate(sys.argv[:-1]):
            if argv == '--locale':
                locale = safe_decode(sys.argv[i + 1])
        try:
            qm = resources[f'locale/{locale}.qm']
        except FileNotFoundError:
            oslogger.warning(f'no translation file found for {locale}')
            pass
        self._locale = locale
        return self._locale

    def set_locale(self, translators):
        """"Sets the application language.

        Parameters
        ----------
        translators: list of QTranslator
            For some reason, the QTranslator object needs to be created in the
            main function. Therefore it is passed as an argument.
        """
        try:
            qm = resources[f'locale/{self.locale}.qm']
        except FileNotFoundError:
            return
        self.translators = translators
        self._main_translator = self.translators.pop()
        self._main_translator.load(qm)
        QtWidgets.QApplication.installTranslator(self._main_translator)

    def set_unsaved(self, unsaved_changes=True):
        """Sets the unsaved changes status.

        Parameters
        ----------
        unsaved_changes : bool, optional
            Indicates if there are unsaved changes.
        """
        self.unsaved_changes = unsaved_changes
        self.window_message()

    def window_message(self, msg=None):
        """Display a message in the window border, including an unsaved
        message indicator.

        Parameters
        ----------
        msg : str, NoneType, optional
            A message, or None to refresh the window message.
        """
        if msg is not None:
            self.window_msg = msg
        flags = ''
        if self.unsaved_changes:
            flags += ' *'
        if self.read_only:
            flags += _(' [read only]')
        self.setWindowTitle(self.window_msg + '%s - OpenSesame' % flags)

    def update_overview_area(self):
        """Refreshes the overview area."""
        self.experiment.build_item_tree()
        item = self.tabwidget.current_item()
        if item is not None:
            self.experiment.items[item].update()

    def update_preferences_tab(self):
        """
        If the preferences tab is open, make sure that its controls are updated
        to match potential changes to the preferences
        """
        w = self.ui.tabwidget.get_widget('__preferences__')
        if w is not None:
            w.set_controls()

    def show_text_in_toolbar(self):
        """
        Set the toolbar style (text/ icons only) based on the menu action status
        """
        if self.ui.action_show_text_in_toolbar.isChecked():
            style = QtCore.Qt.ToolButtonTextUnderIcon
        else:
            style = QtCore.Qt.ToolButtonIconOnly
        self.ui.toolbar_main.setToolButtonStyle(style)

    def toggle_overview(self, dummy=None):
        """
        Set the visibility of the overview area based on the state of the
        toolbar action

        Keyword arguments:
        dummy -- a dummy argument passed by the signal handler (default=None)
        """
        if not self.ui.action_show_overview.isChecked():
            self.ui.dock_overview.setVisible(False)
            return
        self.ui.dock_overview.setVisible(True)

    def toggle_pool(self, make_visible):
        """
        Refresh the file pool

        Keyword arguments:
        make_visible -- an optional boolean that sets the visibility of the file
                                        pool (default = None)
        """
        if make_visible is not None:
            self.ui.action_show_pool.setChecked(make_visible)
        if not self.ui.action_show_pool.isChecked():
            self.ui.dock_pool.setVisible(False)
            return
        self.ui.dock_pool.setVisible(True)
        self.ui.pool_widget.setFocus()
        self.ui.pool_widget.refresh()

    def save_unsaved_changes(self):
        """Checks whether there are unsaved changes. If so, the user can
        choose to discard or save these changes, or to cancel.

        Returns
        -------
        bool
            False if the user has cancelled, True otherwise.
        """
        from libqtopensesame._input.confirmation import confirmation
        if not self.unsaved_changes:
            return True
        resp = confirmation(
            self,
            msg=_('Your experiment contains unsaved changes. Do you want to save your experiment?'),
            title=_('Save changes?'), allow_cancel=True,
            default='cancel').show()
        if resp is None:
            return False
        if resp:
            self.save_file()
        return True

    def closeEvent(self, e):
        """Process a request to close the application.

        Parameters
        ----------
        e : QCloseEvent
        """
        if self.block_close_event:
            e.ignore()
            return
        if not self.save_unsaved_changes():
            e.ignore()
            return
        self.save_state()
        self.experiment.pool.clean_up()
        self.extension_manager.fire('close')
        e.accept()

    def update_recent_files(self):
        """Recreate the list with recent documents"""
        from libqtopensesame.actions import recent_action

        # Add the current path to the front of the list
        if self.current_path is not None and os.path.exists(self.current_path):
            if self.current_path in self.recent_files:
                self.recent_files.remove(self.current_path)
            self.recent_files.insert(0, self.current_path)

        # Trim the list
        self.recent_files = self.recent_files[:5]

        # Build the menu
        self.ui.menu_recent_files.clear()
        if len(self.recent_files) == 0:
            a = QtWidgets.QAction(
                _("(No recent files)"),
                self.ui.menu_recent_files
            )
            a.setDisabled(True)
            self.ui.menu_recent_files.addAction(a)
        else:
            for path in self.recent_files:
                self.ui.menu_recent_files.addAction(
                    recent_action.recent_action(
                        path,
                        self,
                        self.ui.menu_recent_files
                    )
                )

    def open_experiment(self, exp, path=None, add_to_recent=True):
        """Opens an experiment from an Experiment object (as opposed to from 
        disk).
        
        Parameters
        ----------
        exp : Experiment
        path : str or None, optional
            The path that is associated with the experiment or None if the
            experiment does not exist on disk.
        add_to_recent : bool, optional
            Indicates whether the experiment should be added to the recent
            files.
        """
        self.experiment.pool.clean_up()
        self.experiment = exp
        self.experiment.build_item_tree()
        self.ui.itemtree.default_fold_state()
        self.ui.tabwidget.open_general()
        if add_to_recent and path is not None:
            self.current_path = path
            self.read_only = not os.access(path, os.W_OK)
            self.window_message(self.current_path)
            self.update_recent_files()
            cfg.default_logfile_folder = os.path.dirname(self.current_path)
        else:
            self.window_message("New experiment")
            self.current_path = None
        self.set_unsaved(False)
        self.ui.pool_widget.refresh()
        self.extension_manager.fire('open_experiment', path=path)
        self.set_busy(False)
        # Process non-fatal errors
        if exp.items.error_log:
            self.tabwidget.open_markdown(
                _(f'Errors occurred while opening the file:\n\n') +
                '\n\n'.join(
                    [str(exc) for exc in exp.items.error_log]
                ),
                title=_('Error'),
                icon='dialog-error'
            )
            self.window_message("New experiment")
            self.current_path = None

    def open_file(self, dummy=None, path=None, add_to_recent=True):
        """Opens an experiment file.

        Parameters
        ----------
        dummy, optional
            Dummy argument passed by event handler.
        path : str, NoneType, optional
            The path to the file. If None, a file dialog is presented.
        add_to_recent : bool, optional
            Indicates whether the file should be added to the list of recent
            experiments.
        """
        from libopensesame.osexpfile import osexpreader

        if not self.save_unsaved_changes():
            self.ui.tabwidget.open_general()
            return
        if path is None:
            path = QtWidgets.QFileDialog.getOpenFileName(
                self.ui.centralwidget,
                _("Open file"),
                filter=OPEN_FILE_FILTER,
                directory=cfg.file_dialog_path
            )
        # In PyQt5, the QFileDialog.getOpenFileName returns a tuple instead of
        # a string, of which the first position contains the path. check for that
        # here.
        if isinstance(path, tuple):
            path = path[0]
        file_handler = self.extension_manager.provide(
            'file_handler',
            path=path
        )
        if file_handler:
            file_handler(path)
            return
        if path is None or not osexpreader.valid_extension(path):
            return
        self.set_busy()
        self.ui.tabwidget.close_all(avoid_empty=False)
        if add_to_recent:
            cfg.file_dialog_path = os.path.dirname(path)
        try:
            exp = Experiment(self, "Experiment", path,
                             experiment_path=os.path.dirname(path))
        except Exception as e:
            md = _(
                '# Failed to open\n\nFailed to open the file for the '
                'following reason:\n\n- '
            ) + safe_decode(e)
            self.tabwidget.open_markdown(md)
            traceback.print_exc()
            self.console.write(e)
            self.set_busy(False)
            return
        self.open_experiment(exp, path, add_to_recent)

    def set_run_status(self, status):

        self._run_status = status

    def run_status(self):

        return self._run_status

    def save_file(self):
        """Saves the current experiment.

        Parameters
        ----------
        dummy, optional
            A dummy argument passed by the signal handler.
        remember : bool, optional
            Indicates whether the file should be included in the list of recent
            files.
        catch : bool, optional
            Indicates whether exceptions should be caught and displayed in a
            notification.
        """
        if self.current_path is None:
            self.save_file_as()
            return
        self.extension_manager.fire('save_experiment', path=self.current_path)
        # Indicate that we're busy
        self.set_busy(True)
        QtWidgets.QApplication.processEvents()
        # Get ready
        try:
            self.get_ready()
        except OSException as e:
            self.console.write(e)
            self.notify(
                _("The following error occurred while trying to save:<br/>%s")
                % e
            )
            self.set_busy(False)
            return
        # Try to save the experiment if it doesn't exist already
        try:
            self.experiment.save(self.current_path, overwrite=True)
            self.set_busy(False)
        except Exception as e:
            self.console.write(e)
            self.notify(
                _("Failed to save file. Error: %s")
                % safe_decode(e)
            )
            self.set_busy(False)
            return
        self.update_recent_files()
        self.set_unsaved(False)
        self.window_message(self.current_path)
        self.set_busy(False)

    def save_file_as(self):
        """Saves the current experiment after asking for a file name."""
        # Choose a default file name based on the experiment title
        if self.current_path is None:
            cfg.file_dialog_path = os.path.join(
                os.path.dirname(cfg.file_dialog_path),
                self.experiment.syntax.sanitize(
                    self.experiment.var.title,
                    strict=True,
                    allow_vars=False
                )
            )
        else:
            cfg.file_dialog_path = self.current_path
        path = QtWidgets.QFileDialog.getSaveFileName(
            self.ui.centralwidget,
            _('Save as…'),
            directory=cfg.file_dialog_path,
            filter=SAVE_FILE_FILTER
        )
        # In PyQt5, the QFileDialog.getOpenFileName returns a tuple instead of
        # a string, of which the first position contains the path.
        if isinstance(path, tuple):
            path = path[0]
        if not path:
            return
        if not path.lower().endswith('.osexp'):
            path += '.osexp'
        cfg.file_dialog_path = os.path.dirname(path)
        self.current_path = path
        self.read_only = False
        cfg.default_logfile_folder = os.path.dirname(self.current_path)
        self.save_file()

    def regenerate(self, script):
        """Regenerates the current experiment from script, and updates the
        GUI.
        """
        self.extension_manager.fire('prepare_regenerate')
        try:
            exp = Experiment(self, name=self.experiment.var.title,
                             string=script,
                             pool_folder=self.experiment.pool.folder(),
                             experiment_path=self.experiment.experiment_path,
                             resources=self.experiment.resources)
        except OSException as e:
            md = _(
                '# Parsing error\n\nFailed to parse the script for the '
                'following reason:\n\n- '
            ) + e.markdown()
            self.tabwidget.open_markdown(md)
            self.console.write(e)
            return
        self.experiment = exp
        self.tabwidget.close_all()
        self.experiment.build_item_tree()
        self.extension_manager.fire('regenerate')

    def update_resolution(self, width, height):
        """Updates the resolution in a way that preserves display centering.
        This is kind of a quick hack. First generate the script, change the
        resolution in the script and then re-parse the script.

        Parameters
        ----------
        width
            The display width in pixels.
        height
            The display height in pixels.
        """
        oslogger.debug("changing resolution to %d x %d" % (width, height))
        try:
            script = self.experiment.to_string()
        except Exception as e:
            if not isinstance(e, OSException):
                e = OSException(
                    'Failed to change the display resolution',
                    exception=e
                )
            md = _(
                '# Error\n\nFailed to change display resolution for the '
                'following reason:\n\n- '
            ) + e.markdown()
            self.tabwidget.open_markdown(md)
            return
        old_cmd = self.experiment.syntax.create_cmd(
            'set', ['height', self.experiment.var.height])
        new_cmd = self.experiment.syntax.create_cmd(
            'set', ['height', height])
        script = script.replace(old_cmd, new_cmd)
        old_cmd = self.experiment.syntax.create_cmd(
            'set', ['width', self.experiment.var.width])
        new_cmd = self.experiment.syntax.create_cmd('set', ['width', width])
        script = script.replace(old_cmd, new_cmd)
        try:
            tmp = Experiment(self, name=self.experiment.var.title,
                             string=script,
                             pool_folder=self.experiment.pool.folder(),
                             experiment_path=self.experiment.experiment_path,
                             resources=self.experiment.resources)
        except OSException as error:
            self.notify(_("Could not parse script: %s") % error)
            self.edit_script.edit.setText(self.experiment.to_string())
            return
        self.experiment = tmp
        self.ui.tabwidget.close_other()
        self.update_overview_area()
        self.extension_manager.fire('regenerate')

    def get_ready(self):
        """Give all items the opportunity to get ready for running or saving"""
        # Redo the get_ready loop until no items report having done
        # anything
        redo = True
        done = []
        while redo:
            redo = False
            for item in self.experiment.items:
                if item not in done:
                    done.append(item)
                    if self.experiment.items[item].get_ready():
                        oslogger.debug("'%s' did something" % item)
                        redo = True
                        break

    def kill_experiment(self):
        """Tries to kill a running experiment. This is not supported by all
        runners.
        """
        self._runner.kill()
        self.ui.action_kill.setDisabled(True)

    def run_experiment(self, dummy=None, fullscreen=True, quick=False):
        """
        Runs the current experiment.

        Parameters
        ----------
        dummy
            A dummy argument that is passed by signaler.
                                        (default=None)
        fullscreen : bool, optional
            A boolean to indicate whether the window should be fullscreen.
        quick : bool, optional
            A boolean to indicate whether default should be used for the
            log-file and subject number. Mostly useful while testing the
            experiment.
        """
        self.get_ready()
        self.enable(False)
        print('\n')
        oslogger.debug('using %s runner' % cfg.runner)
        self._runner = self.runner_cls(self)
        self._runner.run(fullscreen=fullscreen, quick=quick)
        self.enable(True)
    
    def notify(self, msg, title=None, icon=None, **kwargs):
        self.extension_manager.fire('notify', message=msg, **kwargs)

    @property
    def runner_cls(self):
        """
        Returns
        -------
        BaseRunner
            A runner class.
        """
        runner = self.extension_manager.provide('runner')
        if runner is not None:
            return runner
        from libqtopensesame import runners
        return getattr(runners, '%s_runner' % cfg.runner)

    def run_experiment_in_window(self):
        """Runs the experiment in a window"""
        self.run_experiment(fullscreen=False)

    def run_quick(self):
        """Run the experiment without asking for subject nr and logfile"""
        self.run_experiment(fullscreen=False, quick=True)

    def enable(self, enabled=True):
        """Enable or disable parts of the GUI (i.e. those parts that should be
        disabled when the experiment is running.

        Parameters
        ----------
        enabled : bool
        """
        self.block_close_event = not enabled
        self.ui.dock_overview.setEnabled(enabled)
        self.ui.centralwidget.setEnabled(enabled)
        for action in self.ui.toolbar_main.actions():
            # The kill action should be enabled when the experiment is running
            # and the runner supports killing
            action.setEnabled(
                not enabled and self.runner_cls.supports_kill
                if action.objectName() == 'action_kill'
                else enabled
            )
        self.ui.toolbar_items.setEnabled(enabled)
        self.ui.menubar.setEnabled(enabled)
        self.ui.dock_pool.setEnabled(enabled)
        self.ui.dock_overview.setEnabled(enabled)
        
    def new_window(self):
        """Launches a new instance of the application"""
        subprocess.Popen([sys.executable] + sys.argv)

    def _id(self):
        """
        Returns
        -------
        str
            A unique id string for this instance of OpenSesame. This allows us
            to distinguish between different instances of the program that may
            be running simultaneously.
        """
        _id = safe_decode(
            repr(QtWidgets.QApplication.instance()), enc=self.enc)
        return _id

    @property
    def read_only(self):
        """Getter property for toggling the save action when setting."""
        return self._read_only

    @read_only.setter
    def read_only(self, read_only):
        """Setter property for toggling the save action."""
        self._read_only = read_only
        self.ui.action_save.setEnabled(not read_only)


# Alias for backwards compatibility
qtopensesame = QtOpenSesame



# -*- coding:utf-8 -*-

"""
This file is part of openexp.

openexp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

openexp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with openexp.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import OSException
from qtpy import QtWidgets, QtGui
from libqtopensesame.misc.translate import translation_context
from libopensesame.misc import snake_case
from libqtopensesame.dialogs.text_input import TextInput
_ = translation_context(u'sketchpad', category=u'item')


class BaseElement:

    r"""A base class for the sketchpad-element GUIs."""
    def __init__(self, sketchpad, string=None, properties=None):
        r"""Constructor.

        Parameters
        ----------
        sketchpad
            A sketchpad object.
        type
            sketchpad
        string : str, unicode, NoneType, optional
            An element-definition string.
        properties : dict, NoneType, optional
            A dictionary with element properties. If a dictionary is provided,
            the string keyword is ignored, and a definition string is created
            from the properties dict.
        """
        if properties is not None:
            string = f'draw {snake_case(self.__class__.__name__)}'
            for var, val in properties.items():
                string += f' {var}="{val}"'
        super(base_element, self).__init__(sketchpad, string)
        self.selected = False
        self.highlighted = False
        self.graphics_item = None

    @property
    def main_window(self):
        return self.experiment.main_window

    @property
    def theme(self):
        return self.experiment.main_window.theme

    @property
    def console(self):
        return self.experiment.main_window.console

    def draw(self):
        r"""Draw this element, without redrawing the entire sketchpad."""
        self.graphics_item = super(base_element, self).draw()
        self.graphics_item.element = self
        self.graphics_item.setToolTip(self.to_string())
        self.update()

    def move(self, dx=0, dy=0):
        r"""Moves the element position.

        Parameters
        ----------
        dx, optional
            The horizontal movement.
        dy, optional
            The vertical movement.
        """
        for var, val in self.properties.items():
            if var in [u'x', u'x1', u'x2'] and \
                    isinstance(self.properties[var], (int, float)):
                self.properties[var] += dx
            if var in [u'y', u'y1', u'y2'] and \
                    isinstance(self.properties[var], (int, float)):
                self.properties[var] += dy

    def set_pos(self, x=0, y=0):
        r"""Sets the element position.

        Parameters
        ----------
        x, optional
            The horizontal position.
        y, optional
            The vertical position.
        """
        self.properties[u'x'] = x
        self.properties[u'y'] = y

    def get_pos(self):
        r"""Gets the element position.

        Returns
        -------
        tuple
            An (x,y) tuple.
        """
        if u'x' in self.properties:
            return self.properties[u'x'], self.properties[u'y']
        return self.properties[u'x1'], self.properties[u'y1']

    def update(self):
        r"""Redraws this element."""
        if self.graphics_item is None:
            # Sometimes update is called before draw, and the graphics_item
            # doesn't exist yet.
            return
        if self.highlighted:
            self.graphics_item.setGraphicsEffect(self.highlighted_effect())
        elif self.selected:
            self.graphics_item.setGraphicsEffect(self.selected_effect())
        else:
            self.graphics_item.setGraphicsEffect(None)

    def highlight(self, highlighted=True):
        r"""Sets the highlight status of the element and redraws.

        Parameters
        ----------
        highlighted
            A bool indicating the highlight status.
        """
        self.highlighted = highlighted
        self.update()

    def select(self, selected=True):
        r"""Sets the selected status of the element and redraws.

        Parameters
        ----------
        selected
            A bool indicating the selected status.
        """
        self.selected = selected
        self.update()

    def selected_effect(self):
        r"""Creates the selected effect.

        Returns
        -------
        A QGraphicsEffect object.
        """
        effect = QtWidgets.QGraphicsDropShadowEffect()
        effect.setColor(QtGui.QColor('#00FF00'))
        effect.setBlurRadius(32)
        effect.setOffset(8, 8)
        return effect

    def highlighted_effect(self):
        r"""Creates the highlighted effect.

        Returns
        -------
        A QGraphicsEffect object.
        """
        effect = QtWidgets.QGraphicsOpacityEffect()
        return effect

    def get_property(self, name, _type=str, fallback=None):
        r"""Gets an element property.

        Parameters
        ----------
        name
            The property name.
        _type, optional
            The property type.
        fallback, optional
            A fallback value, in case the value cannot be cast to the requested
            type.

        Returns
        -------
        The property in the specified type, or None if the property doesn't
        exist.
        """
        properties = self.eval_properties()
        if name not in properties:
            return None
        val = properties[name]
        if _type == str:
            return str(val)
        if _type == int:
            try:
                return int(val)
            except ValueError:
                return fallback
        if _type == float:
            try:
                return float(val)
            except ValueError:
                return fallback
        if _type == bool:
            if isinstance(val, str):
                if val in (u'yes', u'1'):
                    return True
                if val in (u'no', u'0'):
                    return False
                return fallback
            return bool(val)
        raise TypeError(f'{name} should be {_type} not {type(val)}')

    def set_property(self, name, val, yes_no=False):
        r"""Sets an element property.

        Parameters
        ----------
        name
            The property name.
        val
            The property value.
        yes_no, optional
            Indicates whether the value should be treated as a bool and recoded
            to yes/ no.
        """
        if name not in self.properties:
            return None
        if yes_no:
            if val:
                val = u'yes'
            else:
                val = u'no'
        self.properties[name] = val

    def script_validator(self, s):
        r"""Validates an element script.

        Parameters
        ----------
        s : str
            An element script.

        Returns
        -------
        bool
            True if the script is valid, False otherwise.
        """
        return u'\n' not in s

    def show_script_edit_dialog(self):
        r"""Shows the script-edit dialog."""
        old_string = self.to_string()
        string = TextInput(self.sketchpad._edit_widget,
                           msg=_(u'Element script'), content=self.to_string(),
                           validator=self.script_validator).get_input()
        if string is None:
            return
        try:
            self.from_string(string)
        except OSException as e:
            self.main_window.notify(e)
            self.console.write(e)
            self.from_string(old_string)
            return
        self.sketchpad.draw()

    def show_edit_dialog(self):
        r"""Shows an edit dialog for the element, typically to edit the element
        script.
        """
        self.show_script_edit_dialog()

    def show_context_menu(self, pos):
        r"""Shows a context menu for the element.

        Parameters
        ----------
        pos : QPoint
        """
        from libqtopensesame._input.popup_menu import popup_menu
        pm = popup_menu(self.main_window, [
            (0, _(u'Edit script'), u'utilities-terminal'),
            (1, _(u'Raise to front'), u'go-top'),
            (2, _(u'Lower to bottom'), u'go-bottom'),
            (3, _(u'Specify polar coordinates'), u'accessories-calculator'),
            (4, _(u'Delete'), u'edit-delete'),
        ])
        resp = pm.show()
        if resp is None:
            return
        if resp == 0:
            self.show_script_edit_dialog()
        elif resp == 1:
            self.properties[u'z_index'] = self.sketchpad.min_z_index() - 1
        elif resp == 2:
            self.properties[u'z_index'] = self.sketchpad.max_z_index() + 1
        elif resp == 3:
            self.show_polar_coordinates_dialog()
        elif resp == 4:
            self.sketchpad.remove_elements([self])
        self.sketchpad.draw()

    @classmethod
    def mouse_press(cls, sketchpad, pos):
        r"""A static method that processes mouse clicks and returns an element
        object when one has been created.

        Parameters
        ----------
        sketchpad
            A sketchpad object.
        pos
            An (x,y) tuple with mouse-click coordinates.

        Returns
        -------
        An element object, or None if no element was created.
        """
        return None

    @classmethod
    def mouse_release(cls, sketchpad, pos):
        r"""A static method that processes mouse releases and returns an
        element object when one has been created.

        Parameters
        ----------
        sketchpad
            A sketchpad object.
        pos
            An (x,y) tuple with mouse-release coordinates.

        Returns
        -------
        An element object, or None if no element was created.
        """
        pass

    @classmethod
    def mouse_move(cls, sketchpad, pos):
        r"""A static method that processes mouse moves, and gives elements the
        opportunity to update their preview.

        Parameters
        ----------
        sketchpad
            A sketchpad object.
        pos
            An (x,y) tuple with mousecoordinates.
        """
        pass

    @classmethod
    def reset(cls, sketchpad):

        pass

    @staticmethod
    def requires_text():
        r"""Indicates whether the element requires text settings.

        Returns
        -------
        bool
        """
        return False

    @staticmethod
    def requires_name():
        r"""Indicates whether the element requires name settings.

        Returns
        -------
        bool
        """
        return True

    @staticmethod
    def requires_penwidth():
        r"""Indicates whether the element requires penwidth settings.

        Returns
        -------
        bool
        """
        return False

    @staticmethod
    def requires_color():
        r"""Indicates whether the element requires color settings.

        Returns
        -------
        bool
        """
        return False

    @staticmethod
    def requires_arrow_head_width():
        r"""Indicates whether the element requires arrow_head_width settings.

        Returns
        -------
        bool
        """
        return False

    @staticmethod
    def requires_arrow_body_width():
        r"""Indicates whether the element requires arrow_body_width settings.

        Returns
        -------
        bool
        """
        return False

    @staticmethod
    def requires_arrow_body_length():
        r"""Indicates whether the element requires arrow_body_length settings.

        Returns
        -------
        bool
        """
        return False

    @staticmethod
    def requires_scale():
        r"""Indicates whether the element requires scale settings.

        Returns
        -------
        bool
        """
        return False

    @staticmethod
    def requires_rotation():
        r"""Indicates whether the element requires rotation settings.

        Returns
        -------
        bool
        """
        return False

    @staticmethod
    def requires_fill():
        r"""Indicates whether the element requires fill settings.

        Returns
        -------
        bool
        """
        return False

    @staticmethod
    def requires_center():
        r"""Indicates whether the element requires center settings.

        Returns
        -------
        bool
        """
        return False

    @staticmethod
    def requires_show_if():
        r"""Indicates whether the element requires show-if settings.

        Returns
        -------
        bool
        """
        return True

    @staticmethod
    def cursor():
        r"""Indicates which cursor should be shown when the element tool is
        selected.

        Returns
        -------
        int
            A Qt::CursorShape value
        """
        return u'cursor-pencil', 3, 28

    def show_polar_coordinates_dialog(self):
        r"""The dialog for calculating cartesian coordinates from polar
        coordinates.
        """
        from libqtopensesame.dialogs.polar_coordinates import polar_coordinates

        d = polar_coordinates(self.sketchpad._edit_widget)
        xy = d.get_xy()
        if xy is None:
            return
        self.properties.update({u'x': xy[0], u'y': xy[1]})
        self.sketchpad.draw()


# Alias for backwards compatibility
base_element = BaseElement
